@@ -61,8 +61,9 @@ enum {
 static int ocserv_conv(int msg_size, const struct pam_message **msg, 
 		struct pam_response **resp, void *uptr)
 {
-struct pam_ctx_st * pctx = uptr;
-unsigned i;
+	struct pam_ctx_st * pctx = uptr;
+	unsigned i;
+	int ret;
 
 	if (msg_size == 0)
 		return PAM_SUCCESS;
@@ -79,8 +80,17 @@ unsigned i;
 			case PAM_TEXT_INFO:
 				syslog(LOG_DEBUG, "PAM-auth conv info: %s", msg[i]->msg);
 
-				str_append_str(&pctx->msg, msg[i]->msg);
-				str_append_data(&pctx->msg, " ", 1);
+				// That should never happen, but also not a big deal if we fail to add message here.
+				// coverity[check_return : FALSE]
+				ret = str_append_str(&pctx->msg, msg[i]->msg);
+				if (ret >= 0)
+					ret = str_append_data(&pctx->msg, " ", 1);
+
+				if (ret < 0) {
+					syslog(LOG_ERR, "Error in memory allocation in PAM");
+					return PAM_BUF_ERR;
+				}
+
 				pctx->sent_msg = 1;
 				break;
 			case PAM_PROMPT_ECHO_OFF:
@@ -93,7 +103,11 @@ unsigned i;
 				}
 
 				if (msg[i]->msg) {
-					str_append_str(&pctx->msg, msg[i]->msg);
+					ret = str_append_str(&pctx->msg, msg[i]->msg);
+					if (ret < 0) {
+						syslog(LOG_ERR, "Error in memory allocation in PAM");
+						return PAM_BUF_ERR;
+					}
 				}
 
 				syslog(LOG_DEBUG, "PAM-auth conv: echo-%s, msg: '%s'", (msg[i]->msg_style==PAM_PROMPT_ECHO_ON)?"on":"off", msg[i]->msg!=NULL?msg[i]->msg:"");
@@ -103,8 +117,13 @@ unsigned i;
 				co_resume();
 				pctx->state = PAM_S_INIT;
 
-				if (pctx->password[0] != 0)
+				if (pctx->password[0] != 0) {
 					pctx->replies[i].resp = strdup(pctx->password);
+					if (pctx->replies[i].resp == NULL) {
+						syslog(LOG_ERR, "Error in memory allocation in PAM");
+						return PAM_BUF_ERR;
+					}
+				}
 				pctx->sent_msg = 0;
 				break;
                 }
@@ -160,7 +179,7 @@ int pret;
 struct pam_ctx_st * pctx;
 
 	if (info->username == NULL || info->username[0] == 0) {
-		syslog(LOG_AUTH,
+		syslog(LOG_NOTICE,
 		       "pam-auth: no username present");
 		return ERR_AUTH_FAIL;
 	}
@@ -175,7 +194,7 @@ struct pam_ctx_st * pctx;
 	pctx->dc.appdata_ptr = pctx;
 	pret = pam_start(PACKAGE, info->username, &pctx->dc, &pctx->ph);
 	if (pret != PAM_SUCCESS) {
-		syslog(LOG_AUTH, "PAM-auth init: %s", pam_strerror(pctx->ph, pret));
+		syslog(LOG_NOTICE, "PAM-auth init: %s", pam_strerror(pctx->ph, pret));
 		goto fail1;
 	}
 
@@ -214,7 +233,7 @@ size_t prompt_hash = 0;
 		co_call(pctx->cr);
 
 		if (pctx->cr_ret != PAM_SUCCESS) {
-			syslog(LOG_AUTH, "PAM-auth pam_auth_msg: %s", pam_strerror(pctx->ph, pctx->cr_ret));
+			syslog(LOG_NOTICE, "PAM-auth pam_auth_msg: %s", pam_strerror(pctx->ph, pctx->cr_ret));
 			return ERR_AUTH_FAIL;
 		}
 	}
@@ -254,7 +273,7 @@ struct pam_ctx_st * pctx = ctx;
 		return -1;
 
 	if (pctx->state != PAM_S_WAIT_FOR_PASS) {
-		syslog(LOG_AUTH, "PAM auth: conversation left in wrong state (%d/expecting %d)", pctx->state, PAM_S_WAIT_FOR_PASS);
+		syslog(LOG_NOTICE, "PAM auth: conversation left in wrong state (%d/expecting %d)", pctx->state, PAM_S_WAIT_FOR_PASS);
 		return ERR_AUTH_FAIL;
 	}
 
@@ -265,7 +284,7 @@ struct pam_ctx_st * pctx = ctx;
 	co_call(pctx->cr);
 
 	if (pctx->cr_ret != PAM_SUCCESS) {
-		syslog(LOG_AUTH, "PAM-auth pam_auth_pass: %s", pam_strerror(pctx->ph, pctx->cr_ret));
+		syslog(LOG_NOTICE, "PAM-auth pam_auth_pass: %s", pam_strerror(pctx->ph, pctx->cr_ret));
 		return ERR_AUTH_FAIL;
 	}
 	
@@ -294,7 +313,7 @@ int pret;
 
 	pret = pam_get_item(pctx->ph, PAM_USER, (const void **)&user);
 	if (pret != PAM_SUCCESS) {
-		/*syslog(LOG_AUTH, "PAM-auth: pam_get_item(PAM_USER): %s", pam_strerror(pctx->ph, pret));*/
+		/*syslog(LOG_NOTICE, "PAM-auth: pam_get_item(PAM_USER): %s", pam_strerror(pctx->ph, pret));*/
 		return -1;
 	}
 	
