@@ -35,6 +35,9 @@
 #include "defs.h"
 #include "common/base64-helper.h"
 
+int saved_argc = 0;
+char **saved_argv = NULL;
+
 const char *_vhost_prefix(const char *name)
 {
 	static char tmp[128];
@@ -119,11 +122,10 @@ const char *cmd_request_to_str(unsigned _cmd)
 		return "ban IP";
 	case CMD_BAN_IP_REPLY:
 		return "ban IP reply";
-
+	case CMD_LATENCY_STATS_DELTA:
+		return "latency stats delta";
 	case CMD_SEC_CLI_STATS:
 		return "sm: worker cli stats";
-	case CMD_SECM_CLI_STATS:
-		return "sm: main cli stats";
 	case CMD_SEC_AUTH_INIT:
 		return "sm: auth init";
 	case CMD_SEC_AUTH_CONT:
@@ -134,8 +136,12 @@ const char *cmd_request_to_str(unsigned _cmd)
 		return "sm: decrypt";
 	case CMD_SEC_SIGN:
 		return "sm: sign";
-	case CMD_SECM_STATS:
-		return "sm: stats";
+	case CMD_SEC_SIGN_DATA:
+		return "sm: sign data";
+	case CMD_SEC_SIGN_HASH:
+		return "sm: sign hash";
+	case CMD_SEC_GET_PK:
+		return "sm: get pk";
 	case CMD_SECM_SESSION_CLOSE:
 		return "sm: session close";
 	case CMD_SECM_SESSION_OPEN:
@@ -146,14 +152,18 @@ const char *cmd_request_to_str(unsigned _cmd)
 		return "sm: ban IP";
 	case CMD_SECM_BAN_IP_REPLY:
 		return "sm: ban IP reply";
-	case CMD_SECM_RELOAD:
-		return "sm: reload";
-	case CMD_SECM_RELOAD_REPLY:
-		return "sm: reload reply";
+	case CMD_SECM_CLI_STATS:
+		return "sm: main cli stats";
 	case CMD_SECM_LIST_COOKIES:
 		return "sm: list cookies";
 	case CMD_SECM_LIST_COOKIES_REPLY:
 		return "sm: list cookies reply";
+	case CMD_SECM_STATS:
+		return "sm: stats";
+	case CMD_SECM_RELOAD:
+		return "sm: reload";
+	case CMD_SECM_RELOAD_REPLY:
+		return "sm: reload reply";
 	default:
 		snprintf(tmp, sizeof(tmp), "unknown (%u)", _cmd);
 		return tmp;
@@ -170,6 +180,8 @@ const char *discon_reason_to_str(unsigned reason)
 		return "unspecified";
 	case REASON_USER_DISCONNECT:
 		return "user disconnected";
+	case REASON_TEMP_DISCONNECT:
+		return "anyconnect client disconnected";
 	case REASON_SERVER_DISCONNECT:
 		return "server disconnected";
 	case REASON_IDLE_TIMEOUT:
@@ -180,12 +192,33 @@ const char *discon_reason_to_str(unsigned reason)
 		return "unspecified error";
 	case REASON_SESSION_TIMEOUT:
 		return "session timeout";
+	case REASON_HEALTH_PROBE:
+		return "TCP health probe";
 	default:
 		snprintf(tmp, sizeof(tmp), "unknown (%u)", reason);
 		return tmp;
 	}
 }
 
+unsigned int discon_reason_to_log_level(unsigned int reason)
+{
+	switch (reason) {
+	case 0:
+	case REASON_ANY:
+	case REASON_HEALTH_PROBE:
+		return LOG_DEBUG;
+	case REASON_USER_DISCONNECT:
+	case REASON_TEMP_DISCONNECT:
+	case REASON_SERVER_DISCONNECT:
+	case REASON_IDLE_TIMEOUT:
+	case REASON_DPD_TIMEOUT:
+	case REASON_ERROR:
+	case REASON_SESSION_TIMEOUT:
+	default:
+		return LOG_INFO;
+	}
+}
+
 ssize_t force_write(int sockfd, const void *buf, size_t len)
 {
 	int left = len;
@@ -278,18 +311,30 @@ ssize_t force_read_timeout(int sockfd, void *buf, size_t len, unsigned sec)
 
 void set_non_block(int fd)
 {
-	int val;
+	int val, ret;
 
 	val = fcntl(fd, F_GETFL, 0);
-	fcntl(fd, F_SETFL, val | O_NONBLOCK);
+	ret = fcntl(fd, F_SETFL, val | O_NONBLOCK);
+	if (ret == -1) {
+		/* We log but we do not fail; this seems to be failing when we test
+		 * on 32-bit mode container over a 64-bit kernel. I believe this is related to that:
+		 * https://patchwork.kernel.org/project/qemu-devel/patch/20200331133536.3328-1-linus.walleij@linaro.org/
+		 */
+		int e = errno;
+		syslog(LOG_ERR, "set_non_block: %s", strerror(e));
+	}
 }
 
 void set_block(int fd)
 {
-	int val;
+	int val, ret;
 
 	val = fcntl(fd, F_GETFL, 0);
-	fcntl(fd, F_SETFL, val & (~O_NONBLOCK));
+	ret = fcntl(fd, F_SETFL, val & (~O_NONBLOCK));
+	if (ret == -1) {
+		int e = errno;
+		syslog(LOG_ERR, "set_non_block: %s", strerror(e));
+	}
 }
 
 ssize_t recv_timeout(int sockfd, void *buf, size_t len, unsigned sec)
@@ -876,3 +921,4 @@ size_t oc_strlcpy(char *dst, char const *src, size_t siz)
 }
 
 #endif
+
