@@ -37,6 +37,7 @@
 #include <acct/radius.h>
 #include <auth/plain.h>
 #include <auth/gssapi.h>
+#include <auth/openidconnect.h>
 #include <auth/common.h>
 #include <sec-mod-sup-config.h>
 #include <sec-mod-acct.h>
@@ -52,9 +53,11 @@
 #include <main.h>
 #include <tlslib.h>
 #include <occtl/ctl.h>
+#include <gnutls/crypto.h>
 #include "common-config.h"
 
 #include <getopt.h>
+#include <snapshot.h>
 
 #define OLD_DEFAULT_CFG_FILE "/etc/ocserv.conf"
 #define DEFAULT_CFG_FILE "/etc/ocserv/ocserv.conf"
@@ -100,7 +103,7 @@ static void check_cfg(vhost_cfg_st *vhost, vhost_cfg_st *defvhost, unsigned sile
 
 #define PREAD_STRING(pool, varname) { \
 	unsigned len = strlen(value); \
-	while(c_isspace(value[len-1])) \
+	while(len > 0 && c_isspace(value[len-1])) \
 		len--; \
 	varname = talloc_strndup(pool, value, len); \
 	}
@@ -132,6 +135,15 @@ static void check_cfg(vhost_cfg_st *vhost, vhost_cfg_st *defvhost, unsigned sile
 		varname++; \
 	}
 
+struct snapshot_t * config_snapshot = NULL;
+
+char ** pam_auth_group_list = NULL;
+char ** gssapi_auth_group_list = NULL;
+char ** plain_auth_group_list = NULL;
+unsigned pam_auth_group_list_size = 0;
+unsigned gssapi_auth_group_list_size = 0;
+unsigned plain_auth_group_list_size = 0;
+
 
 /* Parses the string ::1/prefix, to return prefix
  * and modify the string to contain the network only.
@@ -177,8 +189,27 @@ static auth_types_st avail_auth_types[] =
 #endif
 	{NAME("plain"), &plain_auth_funcs, AUTH_TYPE_PLAIN, plain_get_brackets_string},
 	{NAME("certificate"), NULL, AUTH_TYPE_CERTIFICATE, NULL},
+#ifdef 	SUPPORT_OIDC_AUTH
+	{NAME("oidc"), &oidc_auth_funcs, AUTH_TYPE_OIDC, oidc_get_brackets_string},
+#endif
 };
 
+static void check_for_duplicate_password_auth(struct perm_cfg_st *config, const char *vhostname, unsigned type)
+{
+	unsigned i;
+
+	if (type & AUTH_TYPE_USERNAME_PASS) {
+		for (i=0;i<MAX_AUTH_METHODS;i++) {
+			if (config->auth[i].enabled == 0)
+				break;
+			if (config->auth[i].type & AUTH_TYPE_USERNAME_PASS) {
+				fprintf(stderr, ERRSTR"%s: you cannot mix multiple password authentication methods\n", vhostname);
+				exit(1);
+			}
+		}
+	}
+}
+
 static void figure_auth_funcs(void *pool, const char *vhostname,
 			      struct perm_cfg_st *config, char **auth, unsigned auth_size,
 			      unsigned primary)
@@ -189,6 +220,9 @@ static void figure_auth_funcs(void *pool, const char *vhostname,
 	if (auth == NULL)
 		return;
 
+	if (vhostname[0] == 0)
+		vhostname = "vhost:default";
+
 	if (primary != 0) {
 		/* Set the primary authentication methods */
 		for (j=0;j<auth_size;j++) {
@@ -227,7 +261,7 @@ static void figure_auth_funcs(void *pool, const char *vhostname,
 			}
 			talloc_free(auth[j]);
 		}
-		fprintf(stderr, NOTESTR"%ssetting '%s' as primary authentication method\n", vhostname, config->auth[0].name);
+		fprintf(stderr, NOTESTR"%s: setting '%s' as primary authentication method\n", vhostname, config->auth[0].name);
 	} else {
 		unsigned x = config->auth_methods;
 		/* Append authentication methods (alternative options) */
@@ -241,6 +275,7 @@ static void figure_auth_funcs(void *pool, const char *vhostname,
 					config->auth[x].name = talloc_strdup(pool, avail_auth_types[i].name);
 					fprintf(stderr, NOTESTR"%s: enabling '%s' as authentication method\n", vhostname, avail_auth_types[i].name);
 
+					check_for_duplicate_password_auth(config, vhostname, avail_auth_types[i].type);
 					config->auth[x].amod = avail_auth_types[i].mod;
 					config->auth[x].type |= avail_auth_types[i].type;
 					config->auth[x].enabled = 1;
@@ -447,7 +482,7 @@ static void append_iroutes_from_file(struct cfg_st *config, const char *file)
 	ctx.config = config;
 
 	ret = ini_parse(file, iroutes_handler, &ctx);
-	if (ret < 0)
+	if (ret != 0)
 		return;
 
 	for (j=0;j<config->known_iroutes_size;j++) {
@@ -494,7 +529,9 @@ static void apply_default_conf(vhost_cfg_st *vhost, unsigned reload)
 	}
 
 	vhost->perm_config.config->mobile_idle_timeout = (unsigned)-1;
+#ifdef ENABLE_COMPRESSION
 	vhost->perm_config.config->no_compress_limit = DEFAULT_NO_COMPRESS_LIMIT;
+#endif
 	vhost->perm_config.config->rekey_time = 24*60*60;
 	vhost->perm_config.config->cookie_timeout = DEFAULT_COOKIE_RECON_TIMEOUT;
 	vhost->perm_config.config->auth_timeout = DEFAULT_AUTH_TIMEOUT_SECS;
@@ -672,13 +709,13 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 		if (strncmp(section, "vhost:", 6) != 0) {
 			if (reload == 0)
 				fprintf(stderr, WARNSTR"skipping unknown section '%s'\n", section);
-			return 0;
+			return 1;
 		}
 
 		vname = sanitize_name(ctx->pool, section+6);
 		if (vname == NULL || vname[0] == 0) {
 			fprintf(stderr, ERRSTR"virtual host name is illegal '%s'\n", section+6);
-			exit(1);
+			return 0;
 		}
 
 		/* virtual host */
@@ -706,7 +743,7 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 
 	value = sanitize_config_value(vhost->pool, _value);
 	if (value == NULL)
-		return 0;
+		return 1;
 
 	/* read persistent configuration */
 	if (vhost->auth_init == 0) {
@@ -723,8 +760,10 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 		} else if (strcmp(name, "udp-listen-host") == 0) {
 			PREAD_STRING(pool, vhost->perm_config.udp_listen_host);
 		} else if (strcmp(name, "listen-clear-file") == 0) {
-			if (!PWARN_ON_VHOST_STRDUP(vhost->name, "listen-clear-file", unix_conn_file))
-				PREAD_STRING(pool, vhost->perm_config.unix_conn_file);
+			fprintf(stderr, ERRSTR"the 'listen-clear-file' option was removed in ocserv 1.1.2\n");
+			return 0;
+		} else if (strcmp(name, "listen-netns") == 0) {
+			vhost->perm_config.listen_netns_name = talloc_strdup(pool, value);
 		} else if (strcmp(name, "tcp-port") == 0) {
 			if (!PWARN_ON_VHOST(vhost->name, "tcp-port", port))
 				READ_NUMERIC(vhost->perm_config.port);
@@ -736,7 +775,7 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 				const struct passwd* pwd = getpwnam(value);
 				if (pwd == NULL) {
 					fprintf(stderr, ERRSTR"unknown user: %s\n", value);
-					exit(1);
+					return 0;
 				}
 				vhost->perm_config.uid = pwd->pw_uid;
 			}
@@ -745,7 +784,7 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 				const struct group* grp = getgrnam(value);
 				if (grp == NULL) {
 					fprintf(stderr, ERRSTR"unknown group: %s\n", value);
-					exit(1);
+					return 0;
 				}
 				vhost->perm_config.gid = grp->gr_gid;
 			}
@@ -753,6 +792,8 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 			READ_MULTI_LINE(vhost->perm_config.cert, vhost->perm_config.cert_size);
 		} else if (strcmp(name, "server-key") == 0) {
 			READ_MULTI_LINE(vhost->perm_config.key, vhost->perm_config.key_size);
+		} else if (strcmp(name, "debug-no-secmod-stats") == 0) {
+			READ_TF(vhost->perm_config.debug_no_secmod_stats);
 		} else if (strcmp(name, "dh-params") == 0) {
 			READ_STRING(vhost->perm_config.dh_params_file);
 		} else if (strcmp(name, "pin-file") == 0) {
@@ -761,10 +802,12 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 			READ_STRING(vhost->perm_config.srk_pin_file);
 		} else if (strcmp(name, "ca-cert") == 0) {
 			READ_STRING(vhost->perm_config.ca);
+#if !defined(OCSERV_WORKER_PROCESS)
 		} else if (strcmp(name, "key-pin") == 0) {
 			READ_STRING(vhost->perm_config.key_pin);
 		} else if (strcmp(name, "srk-pin") == 0) {
 			READ_STRING(vhost->perm_config.srk_pin);
+#endif
 		} else if (strcmp(name, "socket-file") == 0) {
 			if (!PWARN_ON_VHOST_STRDUP(vhost->name, "socket-file", socket_file_prefix))
 				PREAD_STRING(pool, vhost->perm_config.socket_file_prefix);
@@ -784,6 +827,9 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 				READ_STATIC_STRING(pid_file);
 			} else if (reload == 0)
 				fprintf(stderr, NOTESTR"skipping 'pid-file' config option\n");
+		} else if (strcmp(name, "sec-mod-scale") == 0) {
+			if (!PWARN_ON_VHOST(vhost->name, "sec-mod-scale", sec_mod_scale))
+				READ_NUMERIC(vhost->perm_config.sec_mod_scale);
 		} else {
 			stage1_found = 0;
 		}
@@ -823,10 +869,15 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 	} else if (strcmp(name, "rate-limit-ms") == 0) {
 		if (!WARN_ON_VHOST(vhost->name, "rate-limit-ms", rate_limit_ms))
 			READ_NUMERIC(config->rate_limit_ms);
+	} else if (strcmp(name, "server-drain-ms") == 0) {
+		if (!WARN_ON_VHOST(vhost->name, "server-drain-ms", server_drain_ms))
+			READ_NUMERIC(config->server_drain_ms);
 	} else if (strcmp(name, "ocsp-response") == 0) {
 		READ_STRING(config->ocsp_response);
+#ifdef ANYCONNECT_CLIENT_COMPAT
 	} else if (strcmp(name, "user-profile") == 0) {
 		READ_STRING(config->xml_config_file);
+#endif 
 	} else if (strcmp(name, "default-domain") == 0) {
 		READ_STRING(config->default_domain);
 	} else if (strcmp(name, "crl") == 0) {
@@ -848,6 +899,8 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 		fprintf(stderr, WARNSTR"the option 'session-control' is deprecated\n");
 	} else if (strcmp(name, "banner") == 0) {
 		READ_STRING(config->banner);
+	} else if (strcmp(name, "pre-login-banner") == 0) {
+		READ_STRING(config->pre_login_banner);
 	} else if (strcmp(name, "dtls-legacy") == 0) {
 		READ_TF(config->dtls_legacy);
 	} else if (strcmp(name, "cisco-client-compat") == 0) {
@@ -863,16 +916,20 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 			READ_TF(config->dtls_psk);
 	} else if (strcmp(name, "match-tls-dtls-ciphers") == 0) {
 		READ_TF(config->match_dtls_and_tls);
+#ifdef ENABLE_COMPRESSION
 	} else if (strcmp(name, "compression") == 0) {
 		READ_TF(config->enable_compression);
 	} else if (strcmp(name, "compression-algo-priority") == 0) {
 		if (!WARN_ON_VHOST_ONLY(vhost->name, "compression-algo-priority")) {
+#if defined(OCSERV_WORKER_PROCESS)
 			if (switch_comp_priority(pool, value) == 0) {
 				fprintf(stderr, WARNSTR"invalid compression modstring %s\n", value);
 			}
+#endif
 		}
 	} else if (strcmp(name, "no-compress-limit") == 0) {
 		READ_NUMERIC(config->no_compress_limit);
+#endif
 	} else if (strcmp(name, "use-seccomp") == 0) {
 		READ_TF(config->isolate);
 		if (config->isolate)
@@ -906,7 +963,7 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 		ret = cfg_parse_ports(pool, &config->fw_ports, &config->n_fw_ports, value);
 		if (ret < 0) {
 			fprintf(stderr, ERRSTR"cannot parse restrict-user-to-ports\n");
-			exit(1);
+			return 0;
 		}
 	} else if (strcmp(name, "tls-priorities") == 0) {
 		READ_STRING(config->priorities);
@@ -935,7 +992,7 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 			config->rekey_method = REKEY_METHOD_NEW_TUNNEL;
 		else {
 			fprintf(stderr, ERRSTR"unknown rekey method '%s'\n", value);
-			exit(1);
+			return 0;
 		}
 	} else if (strcmp(name, "cookie-timeout") == 0) {
 		READ_NUMERIC(config->cookie_timeout);
@@ -998,7 +1055,7 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 
 		if (valid_ipv6_prefix(config->network.ipv6_prefix) == 0) {
 			fprintf(stderr, ERRSTR"invalid IPv6 prefix: %u\n", prefix);
-			exit(1);
+			return 0;
 		}
 	} else if (strcmp(name, "ipv6-subnet-prefix") == 0) {
 		/* read subnet prefix */
@@ -1008,7 +1065,7 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 
 			if (valid_ipv6_prefix(prefix) == 0) {
 				fprintf(stderr, ERRSTR"invalid IPv6 subnet prefix: %u\n", prefix);
-				exit(1);
+				return 0;
 			}
 		}
 	} else if (strcmp(name, "custom-header") == 0) {
@@ -1062,14 +1119,40 @@ static int cfg_ini_handler(void *_ctx, const char *section, const char *name, co
 
  exit:
 	talloc_free(value);
-	return 0;
+	return 1;
 }
 
 enum {
 	CFG_FLAG_RELOAD = (1<<0),
-	CFG_FLAG_SECMOD = (1<<1)
+	CFG_FLAG_SECMOD = (1<<1),
+	CFG_FLAG_WORKER = (1<<2)
 };
 
+static void replace_file_with_snapshot(char ** file_name)
+{
+	char * snapshot_file_name;
+	if (*file_name == NULL) {
+		return;
+	}
+
+	if (snapshot_lookup_filename(
+			config_snapshot, 
+			*file_name, 
+			&snapshot_file_name) < 0) {
+		fprintf(stderr, ERRSTR"cannot find snapshot for file %s\n", *file_name);
+		exit(1);
+	}
+
+	talloc_free(*file_name);
+	*file_name = snapshot_file_name;
+}
+
+#define CONFIG_ERROR(filename, err) { \
+	if (err > 0) \
+		fprintf(stderr, ERRSTR"config file error in line %d\n", err); \
+	else \
+		fprintf(stderr, ERRSTR"cannot load config file %s\n", filename); }
+
 static void parse_cfg_file(void *pool, const char *file, struct list_head *head,
 			   unsigned flags)
 {
@@ -1084,17 +1167,91 @@ static void parse_cfg_file(void *pool, const char *file, struct list_head *head,
 	ctx.reload = (flags&CFG_FLAG_RELOAD)?1:0;
 	ctx.head = head;
 
-	/* parse configuration
-	 */
-	ret = ini_parse(file, cfg_ini_handler, &ctx);
-	if (ret < 0 && file != NULL && strcmp(file, DEFAULT_CFG_FILE) == 0)
-		ret = ini_parse(OLD_DEFAULT_CFG_FILE, cfg_ini_handler, &ctx);
+#if defined(PROC_FS_SUPPORTED)
+	// Worker always reads from snapshot
+	if ((flags & CFG_FLAG_WORKER) == CFG_FLAG_WORKER) {
+		char * snapshot_file = NULL;
 
-	if (ret < 0) {
-		fprintf(stderr, ERRSTR"cannot load config file %s\n", file);
+		if ((snapshot_lookup_filename(config_snapshot, file, &snapshot_file) < 0) && 
+			(snapshot_lookup_filename(config_snapshot, OLD_DEFAULT_CFG_FILE, &snapshot_file) < 0)) {
+			fprintf(stderr, ERRSTR"snapshot_lookup failed for file %s\n", file);
+			exit(1);
+		}
+
+		ret = ini_parse(snapshot_file, cfg_ini_handler, &ctx);
+		if (ret != 0) {
+			CONFIG_ERROR(file, ret);
+			exit(1);
+		}
+		talloc_free(snapshot_file);
+
+		// Walk the config, replacing filename with the snapshot equivalent
+		list_for_each(head, vhost, list) {
+			size_t index;
+			replace_file_with_snapshot(&vhost->perm_config.dh_params_file);
+			replace_file_with_snapshot(&vhost->perm_config.config->ocsp_response);
+			for (index = 0; index < vhost->perm_config.cert_size; index ++) {
+				replace_file_with_snapshot(&vhost->perm_config.cert[index]);
+			}
+		}
+	} else {
+		const char * cfg_file = file;
+
+		if (cfg_file == NULL) {
+			fprintf(stderr, ERRSTR"no config file!\n");
+			exit(1);
+		}
+
+		/* parse configuration
+		*/
+		ret = ini_parse(cfg_file, cfg_ini_handler, &ctx);
+		if (ret < 0 && strcmp(file, DEFAULT_CFG_FILE) == 0) {
+			cfg_file = OLD_DEFAULT_CFG_FILE;
+			ret = ini_parse(cfg_file, cfg_ini_handler, &ctx);
+		}
+
+		if (ret != 0) {
+			CONFIG_ERROR(cfg_file, ret);
+			exit(1);
+		}
+		
+		ret = snapshot_create(config_snapshot, cfg_file);
+		if (ret < 0){
+			fprintf(stderr, ERRSTR"cannot snapshot config file %s\n", cfg_file);
+			exit(1);
+		}
+		list_for_each(head, vhost, list) {
+			size_t index;
+			snapshot_create(config_snapshot, vhost->perm_config.dh_params_file);
+			snapshot_create(config_snapshot, vhost->perm_config.config->ocsp_response);
+			for (index = 0; index < vhost->perm_config.cert_size; index ++) {
+				snapshot_create(config_snapshot, vhost->perm_config.cert[index]);
+			}
+		}
+
+	}
+#else
+	const char * cfg_file = file;
+
+	if (cfg_file == NULL) {
+		fprintf(stderr, ERRSTR"no config file!\n");
 		exit(1);
 	}
 
+	/* parse configuration
+	*/
+	ret = ini_parse(cfg_file, cfg_ini_handler, &ctx);
+	if (ret < 0 && file != NULL && strcmp(file, DEFAULT_CFG_FILE) == 0) {
+		cfg_file = OLD_DEFAULT_CFG_FILE;
+		ret = ini_parse(cfg_file, cfg_ini_handler, &ctx);
+	}
+
+	if (ret != 0) {
+		CONFIG_ERROR(cfg_file, ret);
+		exit(1);
+	}
+#endif
+
 	/* apply configuration not yet applied.
 	 * We start from the last, which is the default server (firstly
 	 * added).
@@ -1118,6 +1275,20 @@ static void parse_cfg_file(void *pool, const char *file, struct list_head *head,
 
 		if (vhost->auto_select_group != 0 && vhost->perm_config.auth[0].amod != NULL && vhost->perm_config.auth[0].amod->group_list != NULL) {
 			vhost->perm_config.auth[0].amod->group_list(config, vhost->perm_config.auth[0].additional, &config->group_list, &config->group_list_size);
+			switch (vhost->perm_config.auth[0].amod->type) {
+			case AUTH_TYPE_PAM|AUTH_TYPE_USERNAME_PASS:
+				pam_auth_group_list = config->group_list;
+				pam_auth_group_list_size = config->group_list_size;
+				break;
+			case AUTH_TYPE_GSSAPI:
+				gssapi_auth_group_list = config->group_list;
+				gssapi_auth_group_list_size = config->group_list_size;
+				break;
+			case AUTH_TYPE_PLAIN|AUTH_TYPE_USERNAME_PASS:
+				plain_auth_group_list = config->group_list;
+				plain_auth_group_list_size = config->group_list_size;
+				break;
+			}
 		}
 
 		if (vhost->expose_iroutes != 0) {
@@ -1176,12 +1347,10 @@ static void check_cfg(vhost_cfg_st *vhost, vhost_cfg_st *defvhost, unsigned sile
 		}
 	}
 
-	if (vhost->perm_config.port == 0 && vhost->perm_config.unix_conn_file == NULL) {
+	if (vhost->perm_config.port == 0) {
 		if (defvhost) {
 			if (vhost->perm_config.port)
-				vhost->perm_config.port = vhost->perm_config.port;
-			else if (vhost->perm_config.unix_conn_file)
-				vhost->perm_config.unix_conn_file = talloc_strdup(vhost, vhost->perm_config.unix_conn_file);
+				vhost->perm_config.port = defvhost->perm_config.port;
 		} else {
 			fprintf(stderr, ERRSTR"%sthe tcp-port option is mandatory!\n", PREFIX_VHOST(vhost));
 			exit(1);
@@ -1245,13 +1414,7 @@ static void check_cfg(vhost_cfg_st *vhost, vhost_cfg_st *defvhost, unsigned sile
 		}
 	}
 
-	if (vhost->perm_config.unix_conn_file != NULL && (config->cert_req != 0)) {
-		if (config->listen_proxy_proto == 0) {
-			fprintf(stderr, ERRSTR"%sthe option 'listen-clear-file' cannot be combined with 'auth=certificate'\n", PREFIX_VHOST(vhost));
-			exit(1);
-		}
-	}
-
+#ifdef ANYCONNECT_CLIENT_COMPAT
 	if (vhost->perm_config.cert && vhost->perm_config.cert_hash == NULL) {
 		vhost->perm_config.cert_hash = calc_sha1_hash(vhost->pool, vhost->perm_config.cert[0], 1);
 	}
@@ -1274,13 +1437,21 @@ static void check_cfg(vhost_cfg_st *vhost, vhost_cfg_st *defvhost, unsigned sile
 			exit(1);
 		}
 	}
+#endif
 
 	if (config->priorities == NULL) {
-		/* on vhosts assign the main host priorities */
+		char *tmp = "";
+		/* on vhosts assign the main host priorities. We furthermore disable TLS1.3 on Cisco clients
+		 * due to issue #318. */
+
+		if (config->cisco_client_compat) {
+			tmp = ":-VERS-TLS1.3";
+		}
+
 		if (defvhost) {
-			config->priorities = talloc_strdup(config, defvhost->perm_config.config->priorities);
+			config->priorities = talloc_asprintf(config, "%s%s", defvhost->perm_config.config->priorities, tmp);
 		} else {
-			config->priorities = talloc_strdup(config, "NORMAL:%SERVER_PRECEDENCE:%COMPAT");
+			config->priorities = talloc_asprintf(config, "%s%s", "NORMAL:%SERVER_PRECEDENCE:%COMPAT", tmp);
 		}
 	}
 
@@ -1309,13 +1480,6 @@ static void check_cfg(vhost_cfg_st *vhost, vhost_cfg_st *defvhost, unsigned sile
 		config->dtls_legacy = 1;
 	}
 
-	if (vhost->perm_config.unix_conn_file) {
-		if (config->dtls_psk && !silent) {
-			fprintf(stderr, NOTESTR"%s'dtls-psk' cannot be combined with unix socket file\n", PREFIX_VHOST(vhost));
-		}
-		config->dtls_psk = 0;
-	}
-
 	if (config->match_dtls_and_tls) {
 		if (config->dtls_legacy) {
 			fprintf(stderr, ERRSTR"%s'match-tls-dtls-ciphers' cannot be applied when 'dtls-legacy' or 'cisco-client-compat' is on\n", PREFIX_VHOST(vhost));
@@ -1326,8 +1490,10 @@ static void check_cfg(vhost_cfg_st *vhost, vhost_cfg_st *defvhost, unsigned sile
 	if (config->mobile_idle_timeout == (unsigned)-1)
 		config->mobile_idle_timeout = config->idle_timeout;
 
+#ifdef ENABLE_COMPRESSION
 	if (config->no_compress_limit < MIN_NO_COMPRESS_LIMIT)
 		config->no_compress_limit = MIN_NO_COMPRESS_LIMIT;
+#endif
 
 	/* use tcp listen host by default */
 	if (vhost->perm_config.udp_listen_host ==  NULL) {
@@ -1414,7 +1580,7 @@ void usage(void)
 	fprintf(stderr, "Please send bug reports to:  "PACKAGE_BUGREPORT"\n");
 }
 
-int cmd_parser (void *pool, int argc, char **argv, struct list_head *head)
+int cmd_parser (void *pool, int argc, char **argv, struct list_head *head, bool worker)
 {
 	unsigned test_only = 0;
 	int c;
@@ -1467,7 +1633,7 @@ int cmd_parser (void *pool, int argc, char **argv, struct list_head *head)
 		exit(1);
 	}
 
-	parse_cfg_file(pool, cfg_file, head, 0);
+	parse_cfg_file(pool, cfg_file, head, worker ? CFG_FLAG_WORKER : 0);
 
 	if (test_only)
 		exit(0);
@@ -1574,6 +1740,9 @@ static void print_version(void)
 	append("PKCS#11");
 #ifdef ANYCONNECT_CLIENT_COMPAT
 	append("AnyConnect");
+#endif
+#ifdef SUPPORT_OIDC_AUTH
+	append("oidc_auth");
 #endif
 	fprintf(stderr, "\n");
 
@@ -1634,7 +1803,7 @@ void remove_pid_file(void)
 	if (pid_file[0]==0)
 		return;
 
-	remove(pid_file);
+	(void)remove(pid_file);
 }
 
 int _add_multi_line_val(void *pool, char ***varname, size_t *num,
@@ -1679,3 +1848,154 @@ void clear_old_configs(struct list_head *head)
 		}
 	}
 }
+
+// ocserv and ocserv-worker both load and parse the configuration files.
+// As part of the process of loading the config files, auth / acct methods 
+// are enabled based on the content of the acct_mod_st and auth_mod_st tables.
+// These auth tables are present in the auth sub-subsystem. Linking against
+// the auth subsystem pulls in a very large set of dependent binaries which
+// increases the overall memory footprint. To avoid this, we provide stub 
+// versions of acct_mod_st and auth_mod_st tables that the ocserv-worker
+// process can link against.
+#if defined(OCSERV_WORKER_PROCESS)
+
+// Group information is populated by the auth subsystem. 
+// When compiles as part of ocserv-worker, the auth subsystem is not present.
+// To work around this, the group information is passed from ocserv-main to
+// ocserv-worker, which then caches it and returns it when queried.
+static void pam_group_list(void *pool, void *_additional, char ***groupname, unsigned *groupname_size)
+{
+	*groupname = pam_auth_group_list;
+	*groupname_size = pam_auth_group_list_size;
+}
+
+static void gssapi_group_list(void *pool, void *_additional, char ***groupname, unsigned *groupname_size)
+{
+	*groupname = gssapi_auth_group_list;
+	*groupname_size = gssapi_auth_group_list_size;
+}
+
+static void plain_group_list(void *pool, void *_additional, char ***groupname, unsigned *groupname_size)
+{
+	*groupname = plain_auth_group_list;
+	*groupname_size = plain_auth_group_list_size;
+}
+
+const struct acct_mod_st radius_acct_funcs = {
+	.type = ACCT_TYPE_RADIUS,
+	.auth_types = ALL_AUTH_TYPES,
+	.vhost_init = NULL,
+	.vhost_deinit = NULL,
+	.open_session = NULL,
+	.close_session = NULL,
+	.session_stats = NULL
+};
+
+const struct acct_mod_st pam_acct_funcs = {
+  .type = ACCT_TYPE_PAM,
+  .auth_types = ALL_AUTH_TYPES,
+  .open_session = NULL,
+  .close_session = NULL,
+};
+
+const struct auth_mod_st pam_auth_funcs = {
+  .type = AUTH_TYPE_PAM | AUTH_TYPE_USERNAME_PASS,
+  .auth_init = NULL,
+  .auth_deinit = NULL,
+  .auth_msg = NULL,
+  .auth_pass = NULL,
+  .auth_group = NULL,
+  .auth_user = NULL,
+  .group_list = pam_group_list
+};
+
+const struct auth_mod_st gssapi_auth_funcs = {
+	.type = AUTH_TYPE_GSSAPI,
+	.auth_init = NULL,
+	.auth_deinit = NULL,
+	.auth_msg = NULL,
+	.auth_pass = NULL,
+	.auth_user = NULL,
+	.auth_group = NULL,
+	.vhost_init = NULL,
+	.vhost_deinit = NULL,
+	.group_list = gssapi_group_list
+};
+
+const struct auth_mod_st plain_auth_funcs = {
+	.type = AUTH_TYPE_PLAIN | AUTH_TYPE_USERNAME_PASS,
+	.allows_retries = 1,
+	.vhost_init = NULL,
+	.auth_init = NULL,
+	.auth_deinit = NULL,
+	.auth_msg = NULL,
+	.auth_pass = NULL,
+	.auth_user = NULL,
+	.auth_group = NULL,
+	.group_list = plain_group_list
+};
+
+
+const struct auth_mod_st radius_auth_funcs = {
+	.type = AUTH_TYPE_RADIUS | AUTH_TYPE_USERNAME_PASS,
+	.allows_retries = 1,
+	.vhost_init = NULL,
+	.vhost_deinit = NULL,
+	.auth_init = NULL,
+	.auth_deinit = NULL,
+	.auth_msg = NULL,
+	.auth_pass = NULL,
+	.auth_user = NULL,
+	.auth_group = NULL,
+	.group_list = NULL
+};
+
+const struct auth_mod_st oidc_auth_funcs = {
+	.type = AUTH_TYPE_OIDC,
+	.allows_retries = 1,
+	.vhost_init = NULL,
+	.vhost_deinit = NULL,
+	.auth_init = NULL,
+	.auth_deinit = NULL,
+	.auth_msg = NULL,
+	.auth_pass = NULL,
+	.auth_user = NULL,
+	.auth_group = NULL,
+	.group_list = NULL
+};
+
+
+#else
+int get_cert_names(struct worker_st * ws, const gnutls_datum_t * raw)
+{
+	return -1;
+}
+#endif
+
+char secmod_socket_file_name_socket_file[_POSIX_PATH_MAX] = {0};
+
+void restore_secmod_socket_file_name(const char * save_path)
+{
+	strlcpy(secmod_socket_file_name_socket_file, save_path, sizeof(secmod_socket_file_name_socket_file));
+}
+
+/* Creates a permanent filename to use for secmod to main communication
+ */
+const char *secmod_socket_file_name(struct perm_cfg_st *perm_config)
+{
+	unsigned int rnd;
+	int ret;
+
+	if (secmod_socket_file_name_socket_file[0] != 0)
+		return secmod_socket_file_name_socket_file;
+
+	ret = gnutls_rnd(GNUTLS_RND_NONCE, &rnd, sizeof(rnd));
+	if (ret < 0)
+		exit(1);
+
+	/* make socket name */
+	snprintf(secmod_socket_file_name_socket_file, sizeof(secmod_socket_file_name_socket_file), "%s.%x",
+		 perm_config->socket_file_prefix, rnd);
+
+	return secmod_socket_file_name_socket_file;
+}
