@@ -43,6 +43,10 @@
 #include <arpa/inet.h>
 #include <ccan/hash/hash.h>
 #include <ccan/htable/htable.h>
+#include <ifaddrs.h>
+#include <sys/socket.h>
+
+static bool if_address_test_local(main_server_st * s, struct sockaddr_storage *addr);
 
 static size_t rehash(const void *_e, void *unused)
 {
@@ -80,7 +84,7 @@ void *main_ban_db_init(main_server_st *s)
 
 void main_ban_db_deinit(main_server_st *s)
 {
-struct htable *db = s->ban_db;
+	struct htable *db = s->ban_db;
 
 	if (db != NULL) {
 		htable_clear(db);
@@ -88,14 +92,27 @@ struct htable *db = s->ban_db;
 	}
 }
 
+#define IS_BANNED(main, entry) (entry->score >= GETCONFIG(main)->max_ban_score)
+
 unsigned main_ban_db_elems(main_server_st *s)
 {
-struct htable *db = s->ban_db;
+	struct htable *db = s->ban_db;
+	ban_entry_st *t;
+	struct htable_iter iter;
+	time_t now = time(0);
+	unsigned banned = 0;
 
-	if (db)
-		return db->elems;
-	else
+	if (db == NULL || GETCONFIG(s)->max_ban_score == 0)
 		return 0;
+
+	t = htable_first(db, &iter);
+	while (t != NULL) {
+		if (t->expires > now && IS_BANNED(s, t)) {
+			banned++;
+		}
+		t = htable_next(db, &iter);
+	}
+	return banned;
 }
 
 static void massage_ipv6_address(ban_entry_st *t)
@@ -155,17 +172,19 @@ int add_ip_to_ban_list(main_server_st *s, const unsigned char *ip, unsigned ip_s
 	 * periodically polls the server */
 	if (e->score < GETCONFIG(s)->max_ban_score) {
 		e->expires = expiration;
-		print_msg = 0;
-	} else
 		print_msg = 1;
-	e->score += score;
+	} else
+		print_msg = 0;
+
+	/* prevent overflow */
+	e->score = (e->score + score) > e->score ? (e->score + score) : (e->score);
 
 	if (ip_size == 4)
 		p_str_ip = inet_ntop(AF_INET, ip, str_ip, sizeof(str_ip));
 	else
 		p_str_ip = inet_ntop(AF_INET6, ip, str_ip, sizeof(str_ip));
 
-	if (GETCONFIG(s)->max_ban_score > 0 && e->score >= GETCONFIG(s)->max_ban_score) {
+	if (GETCONFIG(s)->max_ban_score > 0 && IS_BANNED(s, e)) {
 		if (print_msg && p_str_ip) {
 			mslog(s, NULL, LOG_INFO, "added IP '%s' (with score %d) to ban list, will be reset at: %s", str_ip, e->score, ctime(&e->expires));
 		}
@@ -253,6 +272,13 @@ unsigned check_if_banned(main_server_st *s, struct sockaddr_storage *addr, sockl
 	if (db == NULL || GETCONFIG(s)->max_ban_score == 0)
 		return 0;
 
+	(void)(txt);
+
+	if (if_address_test_local(s, addr)) {
+		mslog(s, NULL, LOG_DEBUG, "Not applying ban to local IP: %s", human_addr2((struct sockaddr*)addr, addr_size, txt, sizeof(txt), 0));
+		return 0;
+	}
+
 	in_size = SA_IN_SIZE(addr_size);
 	if (in_size != 4 && in_size != 16) {
 	    	mslog(s, NULL, LOG_ERR, "unknown address type for %s", human_addr2((struct sockaddr*)addr, addr_size, txt, sizeof(txt), 0));
@@ -303,3 +329,118 @@ void cleanup_banned_entries(main_server_st *s)
 	}
 }
 
+int if_address_init(main_server_st *s)
+{
+	struct ifaddrs * ifaddr = NULL, *ifa;
+	if_address_st * local_if_addresses = NULL;
+	int retval = 0;
+	unsigned count = 0;
+	s->if_addresses_count = 0;
+	s->if_addresses = NULL;
+
+	if (getifaddrs(&ifaddr) < 0) {
+		int err = errno;
+		fprintf(stderr, "Failed to read local if address list: %s", strerror(err));
+		goto cleanup;
+	}
+
+	for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
+		if (ifa->ifa_addr == NULL) {
+			continue;
+		}
+		count ++;
+	}
+	
+	local_if_addresses = talloc_array(s, if_address_st, count);
+	if (local_if_addresses == NULL) {
+		fprintf(stderr, "Failed to allocate");
+		goto cleanup;
+	}
+
+	count = 0;
+
+	for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
+		sa_family_t family;
+		if (ifa->ifa_addr == NULL) {
+			continue;
+		}
+		family = ifa->ifa_addr->sa_family;
+		if (family == AF_INET || family == AF_INET6) {
+			memcpy(&local_if_addresses[count].if_addr, ifa->ifa_addr, sizeof(struct sockaddr));
+			memcpy(&local_if_addresses[count].if_netmask, ifa->ifa_netmask, sizeof(struct sockaddr));
+			count++;
+		}
+	}
+
+	s->if_addresses = local_if_addresses;
+	s->if_addresses_count = count;
+	local_if_addresses = NULL;
+
+	retval = 1;
+
+cleanup:
+	if (ifaddr != NULL)
+		freeifaddrs(ifaddr);
+
+	if (local_if_addresses != NULL)
+		talloc_free(local_if_addresses);
+
+	return retval;
+}
+
+static bool test_local_ipv4(struct sockaddr_in * remote, struct sockaddr_in * local, struct sockaddr_in * network)
+{
+	uint32_t l = local->sin_addr.s_addr & network->sin_addr.s_addr;
+	uint32_t r = remote->sin_addr.s_addr & network->sin_addr.s_addr;
+	if (l != r)
+		return false;
+	else
+		return true;
+}
+
+static bool test_local_ipv6(struct sockaddr_in6 * remote, struct sockaddr_in6 * local, struct sockaddr_in6 * network)
+{
+	unsigned index = 0;
+	
+	for (index = 0; index < 4; index ++) {
+		uint32_t l = local->sin6_addr.s6_addr32[index] & network->sin6_addr.s6_addr32[index];
+		uint32_t r = remote->sin6_addr.s6_addr32[index] & network->sin6_addr.s6_addr32[index];
+		if (l != r) 
+			return false;
+	}
+	return true;
+}
+
+static bool if_address_test_local(main_server_st * s, struct sockaddr_storage *addr)
+{
+	unsigned index;
+	for (index = 0; index < s->if_addresses_count; index ++)
+	{
+		if_address_st * ifa = &s->if_addresses[index];
+		if (ifa->if_addr.sa_family != addr->ss_family)
+			continue;
+
+		switch (addr->ss_family) {
+		case AF_INET:
+			if (test_local_ipv4((struct sockaddr_in *)addr, (struct sockaddr_in *)&ifa->if_addr, (struct sockaddr_in *)&ifa->if_netmask))
+				return true;
+			break;
+		case AF_INET6:
+			if (test_local_ipv6((struct sockaddr_in6 *)addr, (struct sockaddr_in6 *)&ifa->if_addr, (struct sockaddr_in6 *)&ifa->if_netmask))
+				return true;
+			break;
+		default:
+			break;
+		}
+	}
+	return false;
+}
+
+void if_address_cleanup(main_server_st * s)
+{
+	if (s->if_addresses)
+		talloc_free(s->if_addresses);
+
+	s->if_addresses = NULL;
+	s->if_addresses_count = 0;
+}
\ No newline at end of file
