@@ -111,13 +111,13 @@ void remove_proc(main_server_st * s, struct proc_st *proc, unsigned flags)
 
 	/* close any pending sessions */
 	if (proc->active_sid && !(flags & RPROC_QUIT)) {
-		if (session_close(s, proc) < 0) {
+		if (session_close(&(s->sec_mod_instances[proc->sec_mod_instance_index]), proc) < 0) {
 			mslog(s, proc, LOG_ERR, "error closing session (communication with sec-mod issue)");
 			exit(1);
 		}
 	}
 
-	mslog(s, proc, LOG_INFO, "user disconnected (reason: %s, rx: %"PRIu64", tx: %"PRIu64")",
+	mslog(s, proc, discon_reason_to_log_level(proc->discon_reason), "user disconnected (reason: %s, rx: %"PRIu64", tx: %"PRIu64")",
 		discon_reason_to_str(proc->discon_reason), proc->bytes_in, proc->bytes_out);
 
 	pid = remove_from_script_list(s, proc);
