@@ -62,6 +62,12 @@
 #include <grp.h>
 #include <ip-lease.h>
 #include <ccan/list/list.h>
+#include <hmac.h>
+#include <base64-helper.h>
+#include <snapshot.h>
+#include <isolate.h>
+#include <sockdiag.h>
+#include <namespace.h>
 
 #ifdef HAVE_GSSAPI
 # include <libtasn1.h>
@@ -70,24 +76,44 @@ extern const ASN1_ARRAY_TYPE kkdcp_asn1_tab[];
 ASN1_TYPE _kkdcp_pkix1_asn = ASN1_TYPE_EMPTY;
 #endif
 
-int saved_argc = 0;
-char **saved_argv = NULL;
+// Name of environment variable used to pass worker_startup_msg 
+// between ocserv-main and ocserv-worker.
+#define OCSERV_ENV_WORKER_STARTUP_MSG "OCSERV_WORKER_STARTUP_MSG"
+
+extern struct snapshot_t * config_snapshot;
+
+
+int worker_argc = 0;
+char **worker_argv = NULL;
 
 static void listen_watcher_cb (EV_P_ ev_io *w, int revents);
+static void resume_accept_cb (EV_P_ ev_timer *w, int revents);
 
 int syslog_open = 0;
 sigset_t sig_default_set;
 struct ev_loop *loop = NULL;
+static unsigned allow_broken_clients = 0;
+
+typedef struct sec_mod_watcher_st {
+	ev_io sec_mod_watcher;
+	ev_child child_watcher;
+	unsigned int sec_mod_instance_index;
+} sec_mod_watcher_st;
 
 /* EV watchers */
 ev_io ctl_watcher;
-ev_io sec_mod_watcher;
+sec_mod_watcher_st * sec_mod_watchers = NULL;
 ev_timer maintenance_watcher;
+ev_timer graceful_shutdown_watcher;
 ev_signal maintenance_sig_watcher;
 ev_signal term_sig_watcher;
 ev_signal int_sig_watcher;
 ev_signal reload_sig_watcher;
-ev_child child_watcher;
+#if defined(CAPTURE_LATENCY_SUPPORT)
+ev_timer latency_watcher;
+#endif
+
+static bool set_env_from_ws(main_server_st * ws);
 
 static void add_listener(void *pool, struct listen_list_st *list,
 	int fd, int family, int socktype, int protocol,
@@ -107,6 +133,8 @@ static void add_listener(void *pool, struct listen_list_st *list,
 	ev_init(&tmp->io, listen_watcher_cb);
 	ev_io_set(&tmp->io, fd, EV_READ);
 
+	ev_init(&tmp->resume_accept, resume_accept_cb);
+
 	list_add(&list->head, &(tmp->list));
 	list->total++;
 }
@@ -147,8 +175,8 @@ static void set_common_socket_options(int fd)
 }
 
 static 
-int _listen_ports(void *pool, struct perm_cfg_st* config, 
-		struct addrinfo *res, struct listen_list_st *list)
+int _listen_ports(void *pool, struct perm_cfg_st* config, struct addrinfo *res,
+		struct listen_list_st *list, struct netns_fds *netns)
 {
 	struct addrinfo *ptr;
 	int s, y;
@@ -171,8 +199,8 @@ int _listen_ports(void *pool, struct perm_cfg_st* config,
 				type, human_addr(ptr->ai_addr, ptr->ai_addrlen,
 					   buf, sizeof(buf)));
 
-		s = socket(ptr->ai_family, ptr->ai_socktype,
-			   ptr->ai_protocol);
+		s = socket_netns(netns, ptr->ai_family, ptr->ai_socktype,
+				ptr->ai_protocol);
 		if (s < 0) {
 			perror("socket() failed");
 			continue;
@@ -183,8 +211,10 @@ int _listen_ports(void *pool, struct perm_cfg_st* config,
 			y = 1;
 			/* avoid listen on ipv6 addresses failing
 			 * because already listening on ipv4 addresses: */
-			setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
-				   (const void *) &y, sizeof(y));
+			if (setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
+				       (const void *) &y, sizeof(y)) < 0) {
+				perror("setsockopt(IPV6_V6ONLY) failed");
+			}
 		}
 #endif
 
@@ -225,67 +255,12 @@ int _listen_ports(void *pool, struct perm_cfg_st* config,
 	return 0;
 }
 
-static 
-int _listen_unix_ports(void *pool, struct perm_cfg_st* config, 
-		       struct listen_list_st *list)
-{
-	int s, e, ret;
-	struct sockaddr_un sa;
-
-	/* open the UNIX domain socket to accept connections */
-	if (config->unix_conn_file) {
-		memset(&sa, 0, sizeof(sa));
-		sa.sun_family = AF_UNIX;
-		strlcpy(sa.sun_path, config->unix_conn_file, sizeof(sa.sun_path));
-		remove(sa.sun_path);
-
-		if (config->foreground != 0)
-			fprintf(stderr, "listening (UNIX) on %s...\n",
-				sa.sun_path);
-
-		s = socket(AF_UNIX, SOCK_STREAM, 0);
-		if (s == -1) {
-			e = errno;
-			fprintf(stderr, "could not create socket '%s': %s", sa.sun_path,
-			       strerror(e));
-			return -1;
-		}
-
-		umask(006);
-		ret = bind(s, (struct sockaddr *)&sa, SUN_LEN(&sa));
-		if (ret == -1) {
-			e = errno;
-			fprintf(stderr, "could not bind socket '%s': %s", sa.sun_path,
-			       strerror(e));
-			return -1;
-		}
-
-		ret = chown(sa.sun_path, config->uid, config->gid);
-		if (ret == -1) {
-			e = errno;
-			fprintf(stderr, "could not chown socket '%s': %s", sa.sun_path,
-			       strerror(e));
-		}
-
-		ret = listen(s, 1024);
-		if (ret == -1) {
-			e = errno;
-			fprintf(stderr, "could not listen to socket '%s': %s",
-			       sa.sun_path, strerror(e));
-			exit(1);
-		}
-		add_listener(pool, list, s, AF_UNIX, SOCK_TYPE_UNIX, 0, (struct sockaddr *)&sa, sizeof(sa));
-	}
-	fflush(stderr);
-
-	return 0;
-}
-
 /* Returns 0 on success or negative value on error.
  */
 static int
 listen_ports(void *pool, struct perm_cfg_st* config, 
-		struct listen_list_st *list)
+		struct listen_list_st *list,
+		struct netns_fds *netns)
 {
 	struct addrinfo hints, *res;
 	char portname[6];
@@ -367,7 +342,7 @@ listen_ports(void *pool, struct perm_cfg_st* config,
 	}
 #endif
 
-	if (config->port == 0 && config->unix_conn_file == NULL) {
+	if (config->port == 0) {
 		fprintf(stderr, "tcp-port option is mandatory!\n");
 		return -1;
 	}
@@ -390,17 +365,13 @@ listen_ports(void *pool, struct perm_cfg_st* config,
 			return -1;
 		}
 
-		ret = _listen_ports(pool, config, res, list);
+		ret = _listen_ports(pool, config, res, list, netns);
+		freeaddrinfo(res);
+
 		if (ret < 0) {
 			return -1;
 		}
 
-		freeaddrinfo(res);
-	}
-
-	ret = _listen_unix_ports(pool, config, list);
-	if (ret < 0) {
-		return -1;
 	}
 
 	if (list->total == 0) {
@@ -426,7 +397,7 @@ listen_ports(void *pool, struct perm_cfg_st* config,
 			return -1;
 		}
 
-		ret = _listen_ports(pool, config, res, list);
+		ret = _listen_ports(pool, config, res, list, netns);
 		if (ret < 0) {
 			return -1;
 		}
@@ -449,13 +420,17 @@ int y;
 		y = 1;
 		/* avoid listen on ipv6 addresses failing
 		 * because already listening on ipv4 addresses: */
-		setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY,
-			   (const void *) &y, sizeof(y));
+		if (setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY,
+			   (const void *) &y, sizeof(y)) < 0) {
+			perror("setsockopt(IPV6_V6ONLY) failed");
+		}
 	}
 #endif
 
 	y = 1;
-	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const void *) &y, sizeof(y));
+	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (const void *) &y, sizeof(y)) < 0) {
+		perror("setsockopt(SO_REUSEADDR) failed");
+	}
 
 	if (GETCONFIG(s)->try_mtu) {
 		set_mtu_disc(fd, family, 1);
@@ -465,114 +440,12 @@ int y;
 	return;
 }
 
-/* Adjusts the file descriptor limits for the main or worker processes
- */
-static void update_fd_limits(main_server_st *s, unsigned main)
-{
-#ifdef RLIMIT_NOFILE
-	static struct rlimit def_set;
-	struct rlimit new_set;
-	unsigned max;
-	int ret;
-
-	if (main) {
-		ret = getrlimit(RLIMIT_NOFILE, &def_set);
-		if (ret < 0) {
-			fprintf(stderr, "error in getrlimit: %s\n", strerror(errno));
-			exit(1);
-		}
-
-		if (GETCONFIG(s)->max_clients > 0 && GETCONFIG(s)->max_clients > def_set.rlim_cur)
-			max = GETCONFIG(s)->max_clients + 32;
-		else
-			max = MAX(4*1024, def_set.rlim_cur);
-
-		if (max > def_set.rlim_cur) {
-			new_set.rlim_cur = max;
-			new_set.rlim_max = def_set.rlim_max;
-			ret = setrlimit(RLIMIT_NOFILE, &new_set);
-			if (ret < 0) {
-				fprintf(stderr, "error in setrlimit(%u): %s (cur: %u)\n", max, strerror(errno), (unsigned)def_set.rlim_cur);
-			}
-		}
-	} else {
-		/* set limits for worker processes */
-		ret = setrlimit(RLIMIT_NOFILE, &def_set);
-		if (ret < 0) {
-			mslog(s, NULL, LOG_INFO, "cannot update file limit(%u): %s\n", (unsigned)def_set.rlim_cur, strerror(errno));
-		}
-	}
-#endif
-}
-
-
-static void drop_privileges(main_server_st* s)
-{
-	int ret, e;
-	struct rlimit rl;
-
-	if (GETPCONFIG(s)->chroot_dir) {
-		ret = chdir(GETPCONFIG(s)->chroot_dir);
-		if (ret != 0) {
-			e = errno;
-			mslog(s, NULL, LOG_ERR, "cannot chdir to %s: %s", GETPCONFIG(s)->chroot_dir, strerror(e));
-			exit(1);
-		}
-
-		ret = chroot(GETPCONFIG(s)->chroot_dir);
-		if (ret != 0) {
-			e = errno;
-			mslog(s, NULL, LOG_ERR, "cannot chroot to %s: %s", GETPCONFIG(s)->chroot_dir, strerror(e));
-			exit(1);
-		}
-	}
-
-	if (GETPCONFIG(s)->gid != -1 && (getgid() == 0 || getegid() == 0)) {
-		ret = setgid(GETPCONFIG(s)->gid);
-		if (ret < 0) {
-			e = errno;
-			mslog(s, NULL, LOG_ERR, "cannot set gid to %d: %s\n",
-			       (int) GETPCONFIG(s)->gid, strerror(e));
-			exit(1);
-		}
-
-		ret = setgroups(1, &GETPCONFIG(s)->gid);
-		if (ret < 0) {
-			e = errno;
-			mslog(s, NULL, LOG_ERR, "cannot set groups to %d: %s\n",
-			       (int) GETPCONFIG(s)->gid, strerror(e));
-			exit(1);
-		}
-	}
-
-	if (GETPCONFIG(s)->uid != -1 && (getuid() == 0 || geteuid() == 0)) {
-		ret = setuid(GETPCONFIG(s)->uid);
-		if (ret < 0) {
-			e = errno;
-			mslog(s, NULL, LOG_ERR, "cannot set uid to %d: %s\n",
-			       (int) GETPCONFIG(s)->uid, strerror(e));
-			exit(1);
-
-		}
-	}
-
-	update_fd_limits(s, 0);
-
-	rl.rlim_cur = 0;
-	rl.rlim_max = 0;
-	ret = setrlimit(RLIMIT_NPROC, &rl);
-	if (ret < 0) {
-		e = errno;
-		mslog(s, NULL, LOG_ERR, "cannot enforce NPROC limit: %s\n",
-		       strerror(e));
-	}
-}
-
 /* clears the server listen_list and proc_list. To be used after fork().
  * It frees unused memory and descriptors.
  */
 void clear_lists(main_server_st *s)
 {
+	int i;
 	struct listener_st *ltmp = NULL, *lpos;
 	struct proc_st *ctmp = NULL, *cpos;
 	struct script_wait_st *script_tmp = NULL, *script_pos;
@@ -607,13 +480,19 @@ void clear_lists(main_server_st *s)
 	proc_table_deinit(s);
 	ctl_handler_deinit(s);
 	main_ban_db_deinit(s);
+	if_address_cleanup(s);
 
 	/* clear libev state */
 	if (loop) {
 		ev_io_stop (loop, &ctl_watcher);
-		ev_io_stop (loop, &sec_mod_watcher);
-		ev_child_stop (loop, &child_watcher);
+		for (i = 0; i < s->sec_mod_instance_count; i++) {
+			ev_io_stop (loop, &sec_mod_watchers[i].sec_mod_watcher);
+			ev_child_stop (loop, &sec_mod_watchers[i].child_watcher);
+		}
 		ev_timer_stop(loop, &maintenance_watcher);
+#if defined(CAPTURE_LATENCY_SUPPORT)
+		ev_timer_stop(loop, &latency_watcher);		
+#endif
 		/* free memory and descriptors by the event loop */
 		ev_loop_destroy (loop);
 	}
@@ -638,6 +517,7 @@ void clear_lists(main_server_st *s)
 #define TLS_EXT_APP_ID 48018
 #define RECORD_PAYLOAD_POS 13
 #define HANDSHAKE_SESSION_ID_POS 46
+#define HANDSHAKE_RANDOM_POS 14
 
 /* This returns either the application-specific ID extension contents,
  * or the session ID contents. The former is used on the new protocol,
@@ -732,6 +612,30 @@ unsigned get_session_id(main_server_st* s, uint8_t *buffer, size_t buffer_size,
 	return 1;
 }
 
+static
+unsigned has_broken_random(main_server_st* s, uint8_t *buffer, size_t buffer_size)
+{
+	size_t pos,i;
+
+	if (allow_broken_clients)
+		return 0;
+
+	if (buffer_size < RECORD_PAYLOAD_POS+HANDSHAKE_RANDOM_POS+32)
+		return 0;
+
+	/* check whether the client hello contains a random value of all zeros;
+	 * if that's the case it indicates a broken DTLS client. Relates to:
+	 * https://gitlab.com/gnutls/gnutls/-/issues/960 */
+	pos = RECORD_PAYLOAD_POS+HANDSHAKE_RANDOM_POS;
+
+	for (i=0;i<32;i++) {
+		if (buffer[pos+i] != 0)
+			return 0;
+	}
+
+	return 1;
+}
+
 /* A UDP fd will not be forwarded to worker process before this number of
  * seconds has passed. That is to prevent a duplicate message messing the worker.
  */
@@ -765,6 +669,10 @@ int sfd = -1;
 	}
 	buffer_size = ret;
 
+	// Sanitize values returned from oc_recvfrom_at to make coverity happy.
+	cli_addr_size = MIN(sizeof(cli_addr), cli_addr_size);
+	our_addr_size = MIN(sizeof(our_addr), our_addr_size);
+
 	if (buffer_size < RECORD_PAYLOAD_POS) {
 		mslog(s, NULL, LOG_INFO, "%s: too short UDP packet",
 		      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
@@ -797,11 +705,13 @@ int sfd = -1;
 		 * the IP address and forward the socket.
 		 */
 		match_ip_only = 1;
-
-		/* don't bother IP matching when the listen-clear-file is in use */
-		if (GETPCONFIG(s)->unix_conn_file)
-			goto fail;
 	} else {
+		if (has_broken_random(s, s->msg_buffer, buffer_size)) {
+			mslog(s, NULL, LOG_INFO, "%s: detected broken DTLS client hello (no randomness); ignoring",
+			      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
+			goto fail;
+		}
+
 		if (!get_session_id(s, s->msg_buffer, buffer_size, &session_id, &session_id_size)) {
 			mslog(s, NULL, LOG_INFO, "%s: too short handshake packet",
 			      human_addr((struct sockaddr*)&cli_addr, cli_addr_size, tbuf, sizeof(tbuf)));
@@ -827,7 +737,7 @@ int sfd = -1;
 			goto fail;
 		}
 
-		sfd = socket(listener->family, SOCK_DGRAM, listener->protocol);
+		sfd = socket_netns(&s->netns, listener->family, SOCK_DGRAM, listener->protocol);
 		if (sfd < 0) {
 			e = errno;
 			mslog(s, proc_to_send, LOG_ERR, "new UDP socket failed: %s",
@@ -923,7 +833,6 @@ static void sec_mod_child_watcher_cb(struct ev_loop *loop, ev_child *w, int reve
 	ev_child_stop(loop, w);
 	mslog(s, NULL, LOG_ERR, "ocserv-secmod died unexpectedly");
 	ev_feed_signal_event (loop, SIGTERM);
-
 }
 
 void script_child_watcher_cb(struct ev_loop *loop, ev_child *w, int revents)
@@ -970,19 +879,37 @@ static void worker_child_watcher_cb(struct ev_loop *loop, ev_child *w, int reven
 static void kill_children(main_server_st* s)
 {
 	struct proc_st *ctmp = NULL, *cpos;
-
+	int i;
 	/* kill the security module server */
 	list_for_each_safe(&s->proc_list.head, ctmp, cpos, list) {
 		if (ctmp->pid != -1) {
 			remove_proc(s, ctmp, RPROC_KILL|RPROC_QUIT);
 		}
 	}
-	kill(s->sec_mod_pid, SIGTERM);
+
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		kill(s->sec_mod_instances[i].sec_mod_pid, SIGTERM);
+	}
 }
 
-static void term_sig_watcher_cb(struct ev_loop *loop, ev_signal *w, int revents)
+static void kill_children_auth_timeout(main_server_st* s)
+{
+	struct proc_st *ctmp = NULL, *cpos;
+	time_t oldest_permitted_session = time(NULL) - GETCONFIG(s)->auth_timeout;
+
+	/* kill the security module server */
+	list_for_each_safe(&s->proc_list.head, ctmp, cpos, list) {
+		/* If the worker has not completed it's auth within auth_timeout seconds, kill it */
+		if ((ctmp->status < PS_AUTH_COMPLETED) &&
+		    (ctmp->conn_time < oldest_permitted_session) && 
+			(ctmp->pid != -1)) {
+			remove_proc(s, ctmp, RPROC_KILL);
+		}
+	}
+}
+
+static void terminate_server(main_server_st * s)
 {
-	main_server_st *s = ev_userdata(loop);
 	unsigned total = 10;
 
 	mslog(s, NULL, LOG_INFO, "termination request received; waiting for children to die");
@@ -1000,23 +927,61 @@ static void term_sig_watcher_cb(struct ev_loop *loop, ev_signal *w, int revents)
 	ev_break (loop, EVBREAK_ALL);
 }
 
+static void graceful_shutdown_watcher_cb(EV_P_ ev_timer *w, int revents)
+{
+	main_server_st *s = ev_userdata(loop);
+
+	terminate_server(s);
+}
+
+static void term_sig_watcher_cb(struct ev_loop *loop, ev_signal *w, int revents)
+{
+	main_server_st *s = ev_userdata(loop);
+	struct listener_st *ltmp = NULL, *lpos;
+	unsigned int server_drain_ms = GETCONFIG(s)->server_drain_ms;
+
+	if (server_drain_ms == 0) {
+		terminate_server(s);
+	}
+	else 
+	{
+		if (!ev_is_active(&graceful_shutdown_watcher)) {
+			mslog(s, NULL, LOG_INFO, "termination request received; stopping new connections");
+			graceful_shutdown_watcher.repeat = ((ev_tstamp)(server_drain_ms)) / 1000.;
+			mslog(s, NULL, LOG_INFO, "termination request received; waiting %d ms", server_drain_ms);
+			ev_timer_again(loop, &graceful_shutdown_watcher);
+
+			// Close the listening ports and stop the IO
+			list_for_each_safe(&s->listen_list.head, ltmp, lpos, list) {
+				ev_io_stop(loop, &ltmp->io);
+				close(ltmp->fd);
+				list_del(&ltmp->list);
+				talloc_free(ltmp);
+				s->listen_list.total--;
+			}
+		}
+	}
+}
+
 static void reload_sig_watcher_cb(struct ev_loop *loop, ev_signal *w, int revents)
 {
 	main_server_st *s = ev_userdata(loop);
 	int ret;
+	int i;
 
 	mslog(s, NULL, LOG_INFO, "reloading configuration");
-	kill(s->sec_mod_pid, SIGHUP);
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		kill(s->sec_mod_instances[i].sec_mod_pid, SIGHUP);
 
-	/* Reload on main needs to happen later than sec-mod.
-	 * That's because of a test that the certificate matches the
-	 * used key. */
-	ret = secmod_reload(s);
-	if (ret < 0) {
-		mslog(s, NULL, LOG_ERR, "could not reload sec-mod!\n");
-		ev_feed_signal_event (loop, SIGTERM);
+		/* Reload on main needs to happen later than sec-mod.
+		* That's because of a test that the certificate matches the
+		* used key. */
+		ret = secmod_reload(&s->sec_mod_instances[i]);
+		if (ret < 0) {
+			mslog(s, NULL, LOG_ERR, "could not reload sec-mod!\n");
+			ev_feed_signal_event (loop, SIGTERM);
+		}
 	}
-
 	reload_cfg_file(s->config_pool, s->vconfig, 0);
 }
 
@@ -1033,6 +998,21 @@ static void cmd_watcher_cb (EV_P_ ev_io *w, int revents)
 	}
 }
 
+static void resume_accept_cb (EV_P_ ev_timer *w, int revents)
+{
+	main_server_st *s = ev_userdata(loop);
+	struct listener_st *ltmp = (struct listener_st *)((char*)w - offsetof(struct listener_st, resume_accept));
+	// Add hysteresis to the pause/resume cycle to damp oscillations
+	unsigned int resume_threshold = GETCONFIG(s)->max_clients * 9 / 10;
+
+	// Only resume accepting connections if we are under the limit
+	if (resume_threshold == 0 || s->stats.active_clients < resume_threshold) {
+		// Clear the timer and resume accept
+		ev_timer_stop(loop, &ltmp->resume_accept);
+		ev_io_start(loop, &ltmp->io);
+	}
+}
+
 static void listen_watcher_cb (EV_P_ ev_io *w, int revents)
 {
 	main_server_st *s = ev_userdata(loop);
@@ -1042,6 +1022,9 @@ static void listen_watcher_cb (EV_P_ ev_io *w, int revents)
 	int fd, ret;
 	int cmd_fd[2];
 	pid_t pid;
+	int i;
+	hmac_component_st hmac_components[3];
+	char worker_path[_POSIX_PATH_MAX];
 
 	if (ltmp->sock_type == SOCK_TYPE_TCP || ltmp->sock_type == SOCK_TYPE_UNIX) {
 		/* connection on TCP port */
@@ -1094,6 +1077,7 @@ static void listen_watcher_cb (EV_P_ ev_io *w, int revents)
 
 		pid = fork();
 		if (pid == 0) {	/* child */
+			unsigned int sec_mod_instance_index;
 			/* close any open descriptors, and erase
 			 * sensitive data before running the worker
 			 */
@@ -1101,15 +1085,24 @@ static void listen_watcher_cb (EV_P_ ev_io *w, int revents)
 			close(cmd_fd[0]);
 			clear_lists(s);
 			if (s->top_fd != -1) close(s->top_fd);
-			close(s->sec_mod_fd);
-			close(s->sec_mod_fd_sync);
+			for (i = 0; i < s->sec_mod_instance_count; i ++) {
+				close(s->sec_mod_instances[i].sec_mod_fd);
+				close(s->sec_mod_instances[i].sec_mod_fd_sync);
+			}
 
 			setproctitle(PACKAGE_NAME"-worker");
 			kill_on_parent_kill(SIGTERM);
 
+			set_self_oom_score_adj(s);
+
+			sec_mod_instance_index = hash_any(
+				SA_IN_P_GENERIC(&ws->remote_addr, ws->remote_addr_len),
+				SA_IN_SIZE(ws->remote_addr_len), 0) % s->sec_mod_instance_count;
+
 			/* write sec-mod's address */
-			memcpy(&ws->secmod_addr, &s->secmod_addr, s->secmod_addr_len);
-			ws->secmod_addr_len = s->secmod_addr_len;
+			memcpy(&ws->secmod_addr, &s->sec_mod_instances[sec_mod_instance_index].secmod_addr, s->sec_mod_instances[sec_mod_instance_index].secmod_addr_len);
+			ws->secmod_addr_len = s->sec_mod_instances[sec_mod_instance_index].secmod_addr_len;
+
 
 			ws->main_pool = s->main_pool;
 
@@ -1117,26 +1110,56 @@ static void listen_watcher_cb (EV_P_ ev_io *w, int revents)
 
 			ws->cmd_fd = cmd_fd[1];
 			ws->tun_fd = -1;
-			ws->dtls_tptr.fd = -1;
+			set_cloexec_flag(fd, false);
 			ws->conn_fd = fd;
 			ws->conn_type = stype;
+			ws->session_start_time = time(0);
 
-			/* Drop privileges after this point */
-			drop_privileges(s);
+			human_addr2((const struct sockaddr *)&ws->remote_addr, ws->remote_addr_len, ws->remote_ip_str, sizeof(ws->remote_ip_str), 0);
+			human_addr2((const struct sockaddr *)&ws->our_addr, ws->our_addr_len, ws->our_ip_str, sizeof(ws->our_ip_str), 0);
 
-			/* creds and config are not allocated
-			 * under s.
-			 */
-			talloc_free(s);
-#ifdef HAVE_MALLOC_TRIM
-			/* try to return all the pages we've freed to
-			 * the operating system, to prevent the child from
-			 * accessing them. That's totally unreliable, so
-			 * sensitive data have to be overwritten anyway. */
-			malloc_trim(0);
+			hmac_components[0].data = ws->remote_ip_str;
+			hmac_components[0].length = strlen(ws->remote_ip_str);
+			hmac_components[1].data = ws->our_ip_str;
+			hmac_components[1].length = strlen(ws->our_ip_str);
+			hmac_components[2].data = &ws->session_start_time;
+			hmac_components[2].length = sizeof(ws->session_start_time);
+
+			generate_hmac(sizeof(s->hmac_key), s->hmac_key, sizeof(hmac_components) / sizeof(hmac_components[0]), hmac_components, (uint8_t*) ws->sec_auth_init_hmac);
+			
+			// Clear the HMAC key
+			safe_memset((uint8_t*)s->hmac_key, 0, sizeof(s->hmac_key));
+
+			if (!set_env_from_ws(s))
+				exit(1);
+
+#if defined(PROC_FS_SUPPORTED)
+			{
+				char path[_POSIX_PATH_MAX];
+				size_t path_length;
+				path_length = readlink("/proc/self/exe", path, sizeof(path)-1);
+				if (path_length == -1) {
+					mslog(s, NULL, LOG_ERR, "readlink failed %s", strerror(ret));
+					exit(1);
+				}
+				path[path_length] = '\0';
+				if (snprintf(worker_path, sizeof(worker_path), "%s-worker", path) >= sizeof(worker_path)) {
+					mslog(s, NULL, LOG_ERR, "snprint of path %s and ocserv-worker failed", path);
+					exit(1);
+				}
+			}
+#else
+			if (snprintf(worker_path, sizeof(worker_path), "%s-worker", worker_argv[0]) >= sizeof(worker_path)) {
+				mslog(s, NULL, LOG_ERR, "snprint of path %s and ocserv-worker failed", worker_argv[0]);
+				exit(1);
+			}
 #endif
-			vpn_server(ws);
-			exit(0);
+
+			worker_argv[0] = worker_path;
+			execv(worker_path, worker_argv);
+			ret = errno;
+			mslog(s, NULL, LOG_ERR, "exec %s failed %s", worker_path, strerror(ret));
+			exit(1);
 		} else if (pid == -1) {
 fork_failed:
 			mslog(s, NULL, LOG_ERR, "fork failed");
@@ -1165,19 +1188,43 @@ fork_failed:
 		forward_udp_to_owner(s, ltmp);
 	}
 
-	if (GETCONFIG(s)->rate_limit_ms > 0)
-		ms_sleep(GETCONFIG(s)->rate_limit_ms);
+	if (GETCONFIG(s)->max_clients > 0 && s->stats.active_clients >= GETCONFIG(s)->max_clients) {
+		ltmp->resume_accept.repeat = ((ev_tstamp)(1));
+		ev_io_stop(loop, &ltmp->io);
+		ev_timer_again(loop, &ltmp->resume_accept);
+	}
+
+	// Rate limiting of incoming connections is implemented as follows:
+	// After accepting a client connection:
+	//   Arm the flow control timer.
+	//   Stop accepting connections.
+	// When the timer fires, it resumes accepting the connections.
+	if (GETCONFIG(s)->rate_limit_ms > 0) {
+		int rqueue = 0;
+		int wqueue = 0;
+		int retval = sockdiag_query_unix_domain_socket_queue_length(s->sec_mod_instances[0].secmod_addr.sun_path, &rqueue, &wqueue);
+		mslog(s, NULL, LOG_DEBUG, "queue_length retval:%d rqueue:%d wqueue:%d", retval, rqueue, wqueue);
+		if (retval || rqueue > wqueue / 2) {
+			mslog(s, NULL, LOG_INFO, "delaying accepts for %d ms", GETCONFIG(s)->rate_limit_ms);
+			// Arm the timer and pause accept
+			ltmp->resume_accept.repeat = ((ev_tstamp)(GETCONFIG(s)->rate_limit_ms)) / 1000.;
+			ev_io_stop(loop, &ltmp->io);
+			ev_timer_again(loop, &ltmp->resume_accept);
+		}
+	}
 }
 
 static void sec_mod_watcher_cb (EV_P_ ev_io *w, int revents)
 {
+	sec_mod_watcher_st *sec_mod = (sec_mod_watcher_st *)w;
 	main_server_st *s = ev_userdata(loop);
 	int ret;
 
-	ret = handle_sec_mod_commands(s);
+	ret = handle_sec_mod_commands(&s->sec_mod_instances[sec_mod->sec_mod_instance_index]);
 	if (ret < 0) { /* bad commands from sec-mod are unacceptable */
 		mslog(s, NULL, LOG_ERR,
 		       "error in command from sec-mod");
+		ev_io_stop(loop, w);
 		ev_feed_signal_event (loop, SIGTERM);
 	}
 }
@@ -1194,9 +1241,11 @@ static void perform_maintenance(main_server_st *s)
 	vhost_cfg_st *vhost = NULL;
 
 	/* Check if we need to expire any data */
-	mslog(s, NULL, LOG_DEBUG, "performing maintenance (banned IPs: %d)", main_ban_db_elems(s));
+	mslog(s, NULL, LOG_DEBUG, "performing maintenance");
 	cleanup_banned_entries(s);
 	clear_old_configs(s->vconfig);
+	
+	kill_children_auth_timeout(s);
 
 	list_for_each_rev(s->vconfig, vhost, list) {
 		tls_reload_crl(s, vhost, 0);
@@ -1210,6 +1259,25 @@ static void maintenance_watcher_cb(EV_P_ ev_timer *w, int revents)
 	perform_maintenance(s);
 }
 
+#if defined(CAPTURE_LATENCY_SUPPORT)
+static void latency_watcher_cb(EV_P_ ev_timer *w, int revents)
+{
+	main_server_st *s = ev_userdata(loop);
+	s->stats.current_latency_stats = s->stats.delta_latency_stats;
+	s->stats.delta_latency_stats.median_total = 0;
+	s->stats.delta_latency_stats.rms_total = 0;
+	s->stats.delta_latency_stats.sample_count = 0;
+	mslog(
+		s, 
+		NULL, 
+		LOG_DEBUG, 
+		"Latency: Median Total %ld RMS Total %ld Sample Count %ld", 
+		s->stats.current_latency_stats.median_total, 
+		s->stats.current_latency_stats.rms_total, 
+		s->stats.current_latency_stats.sample_count);
+}
+#endif
+
 static void maintenance_sig_watcher_cb(struct ev_loop *loop, ev_signal *w, int revents)
 {
 	main_server_st *s = ev_userdata(loop);
@@ -1227,6 +1295,8 @@ static void syserr_cb (const char *msg)
 	abort();
 }
 
+extern char secmod_socket_file_name_socket_file[_POSIX_PATH_MAX];
+
 int main(int argc, char** argv)
 {
 	int e;
@@ -1236,6 +1306,9 @@ int main(int argc, char** argv)
 	void *worker_pool;
 	void *main_pool, *config_pool;
 	main_server_st *s;
+	char *str;
+	int i;
+	int processor_count = 0;
 
 #ifdef DEBUG_LEAKS
 	talloc_enable_leak_report_full();
@@ -1244,6 +1317,8 @@ int main(int argc, char** argv)
 	saved_argc = argc;
 	saved_argv = argv;
 
+	processor_count = sysconf(_SC_NPROCESSORS_ONLN);
+
 	/* main pool */
 	main_pool = talloc_init("main");
 	if (main_pool == NULL) {
@@ -1257,6 +1332,11 @@ int main(int argc, char** argv)
 		exit(1);
 	}
 
+	if (snapshot_init(config_pool, &config_snapshot, "/tmp/ocserv_") < 0) {
+		fprintf(stderr, "failed to init snapshot");
+		exit(1);
+	}
+
 	s = talloc_zero(main_pool, main_server_st);
 	if (s == NULL) {
 		fprintf(stderr, "memory error\n");
@@ -1267,12 +1347,45 @@ int main(int argc, char** argv)
 	s->stats.start_time = s->stats.last_reset = time(0);
 	s->top_fd = -1;
 	s->ctl_fd = -1;
+	s->netns.default_fd = -1;
+	s->netns.listen_fd = -1;
+
+	if (!hmac_init_key(sizeof(s->hmac_key), (uint8_t*)(s->hmac_key))) {
+		fprintf(stderr, "unable to generate hmac key\n");
+		exit(1);
+	}
+
+	// getopt processing mutates argv. Save a copy to pass to the child.
+	worker_argc = argc;
+	worker_argv = talloc_zero_array(main_pool, char*, worker_argc + 1);
+	if (!worker_argv) {
+		fprintf(stderr, "memory error\n");
+		exit(1);
+	}
+	for (i = 0; i < argc; i ++) {
+		worker_argv[i] = talloc_strdup(main_pool, argv[i]);
+		if (!worker_argv[i]) {
+			fprintf(stderr, "memory error\n");
+			exit(1);
+		}
+	}
+
+	init_fd_limits_default(s);
+
+	str = getenv("OCSERV_ALLOW_BROKEN_CLIENTS");
+	if (str && str[0] == '1' && str[1] == 0)
+		allow_broken_clients = 1;
 
 	list_head_init(&s->proc_list.head);
 	list_head_init(&s->script_list.head);
 	ip_lease_init(&s->ip_leases);
 	proc_table_init(s);
 	main_ban_db_init(s);
+	if (if_address_init(s) == 0)
+	{
+		fprintf(stderr, "failed to initialize local addresses\n");
+		exit(1);
+	}
 
 	sigemptyset(&sig_default_set);
 
@@ -1289,7 +1402,7 @@ int main(int argc, char** argv)
 	}
 	list_head_init(s->vconfig);
 
-	ret = cmd_parser(config_pool, argc, argv, s->vconfig);
+	ret = cmd_parser(config_pool, argc, argv, s->vconfig, false);
 	if (ret < 0) {
 		fprintf(stderr, "Error in arguments\n");
 		exit(1);
@@ -1302,8 +1415,12 @@ int main(int argc, char** argv)
 		exit(1);
 	}
 
+	if (GETPCONFIG(s)->listen_netns_name && open_namespaces(&s->netns, GETPCONFIG(s)) < 0) {
+		fprintf(stderr, "cannot init listen namespaces\n");
+		exit(1);
+	}
 	/* Listen to network ports */
-	ret = listen_ports(s, GETPCONFIG(s), &s->listen_list);
+	ret = listen_ports(s, GETPCONFIG(s), &s->listen_list, &s->netns);
 	if (ret < 0) {
 		fprintf(stderr, "Cannot listen to specified ports\n");
 		exit(1);
@@ -1338,7 +1455,30 @@ int main(int argc, char** argv)
 
 	write_pid_file();
 
-	s->sec_mod_fd = run_sec_mod(s, &s->sec_mod_fd_sync);
+	// Start the configured number of ocserv-sm processes
+	s->sec_mod_instance_count = GETPCONFIG(s)->sec_mod_scale;
+	
+	if (s->sec_mod_instance_count == 0)	{
+		if (GETCONFIG(s)->max_clients != 0) {
+			// Compute ideal number of clients per sec-mod
+			unsigned int sec_mod_count_for_users = GETCONFIG(s)->max_clients / MINIMUM_USERS_PER_SEC_MOD + 1;
+			// Limit it to number of processors. 
+			s->sec_mod_instance_count = MIN(processor_count,sec_mod_count_for_users);
+		} else {
+			// If it's unlimited, the use processor count.
+			s->sec_mod_instance_count = processor_count;
+		}
+	}
+
+	s->sec_mod_instances = talloc_zero_array(s, sec_mod_instance_st, s->sec_mod_instance_count);
+	sec_mod_watchers = talloc_zero_array(s, sec_mod_watcher_st, s->sec_mod_instance_count);
+
+	mslog(s, NULL, LOG_INFO, "Starting %d instances of ocserv-sm", s->sec_mod_instance_count);
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		s->sec_mod_instances[i].server = s;
+		run_sec_mod(&s->sec_mod_instances[i], i);
+	}
+
 	ret = ctl_handler_init(s);
 	if (ret < 0) {
 		mslog(s, NULL, LOG_ERR, "Cannot create command handler");
@@ -1364,12 +1504,14 @@ int main(int argc, char** argv)
 	}
 	ms_sleep(100); /* give some time for sec-mod to initialize */
 
-	s->secmod_addr.sun_family = AF_UNIX;
-	p = s->socket_file;
-	if (GETPCONFIG(s)->chroot_dir) /* if we are on chroot make the socket file path relative */
-		while (*p == '/') p++;
-	strlcpy(s->secmod_addr.sun_path, p, sizeof(s->secmod_addr.sun_path));
-	s->secmod_addr_len = SUN_LEN(&s->secmod_addr);
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		s->sec_mod_instances[i].secmod_addr.sun_family = AF_UNIX;
+		p = s->sec_mod_instances[i].socket_file;
+		if (GETPCONFIG(s)->chroot_dir) /* if we are on chroot make the socket file path relative */
+			while (*p == '/') p++;
+		strlcpy(s->sec_mod_instances[i].secmod_addr.sun_path, p, sizeof(s->sec_mod_instances[i].secmod_addr.sun_path));
+		s->sec_mod_instances[i].secmod_addr_len = SUN_LEN(&s->sec_mod_instances[i].secmod_addr);
+	}
 
 	/* initialize memory for worker process */
 	worker_pool = talloc_named(main_pool, 0, "worker");
@@ -1393,6 +1535,8 @@ int main(int argc, char** argv)
 	}
 #endif
 
+	init_fd_limits_default(s);
+
 	/* increase the number of our allowed file descriptors */
 	update_fd_limits(s, 1);
 
@@ -1400,7 +1544,10 @@ int main(int argc, char** argv)
 	ev_set_syserr_cb(syserr_cb);
 
 	ev_init(&ctl_watcher, ctl_watcher_cb);
-	ev_init(&sec_mod_watcher, sec_mod_watcher_cb);
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		ev_init(&sec_mod_watchers[i].sec_mod_watcher, sec_mod_watcher_cb);
+		sec_mod_watchers[i].sec_mod_instance_index = i;
+	}
 
 	ev_init (&int_sig_watcher, term_sig_watcher_cb);
 	ev_signal_set (&int_sig_watcher, SIGINT);
@@ -1421,19 +1568,32 @@ int main(int argc, char** argv)
 		ev_io_start (loop, &ltmp->io);
 	}
 
-	ev_io_set(&sec_mod_watcher, s->sec_mod_fd, EV_READ);
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		ev_io_set(&sec_mod_watchers[i].sec_mod_watcher, s->sec_mod_instances[i].sec_mod_fd, EV_READ);
+		ev_io_start (loop, &sec_mod_watchers[i].sec_mod_watcher);
+	}
+
 	ctl_handler_set_fds(s, &ctl_watcher);
 
 	ev_io_start (loop, &ctl_watcher);
-	ev_io_start (loop, &sec_mod_watcher);
 
-	ev_child_init(&child_watcher, sec_mod_child_watcher_cb, s->sec_mod_pid, 0);
-	ev_child_start (loop, &child_watcher);
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		ev_child_init(&sec_mod_watchers[i].child_watcher, sec_mod_child_watcher_cb, s->sec_mod_instances[i].sec_mod_pid, 0);
+		ev_child_start (loop, &sec_mod_watchers[i].child_watcher);
+	}
 
 	ev_init(&maintenance_watcher, maintenance_watcher_cb);
 	ev_timer_set(&maintenance_watcher, MAIN_MAINTENANCE_TIME, MAIN_MAINTENANCE_TIME);
 	ev_timer_start(loop, &maintenance_watcher);
 
+	ev_init(&graceful_shutdown_watcher, graceful_shutdown_watcher_cb);
+
+#if defined(CAPTURE_LATENCY_SUPPORT)
+	ev_init(&latency_watcher, latency_watcher_cb);
+	ev_timer_set(&latency_watcher, LATENCY_AGGREGATION_TIME, LATENCY_AGGREGATION_TIME);
+	ev_timer_start(loop, &latency_watcher);
+#endif
+
 	/* allow forcing maintenance with SIGUSR2 */
 	ev_init (&maintenance_sig_watcher, maintenance_sig_watcher_cb);
 	ev_signal_set (&maintenance_sig_watcher, SIGUSR2);
@@ -1445,10 +1605,18 @@ int main(int argc, char** argv)
 	/* try to clean-up everything allocated to ease checks 
 	 * for memory leaks.
 	 */
-	remove(s->full_socket_file);
+	for (i = 0; i < s->sec_mod_instance_count; i ++) {
+		remove(s->sec_mod_instances[i].full_socket_file);
+	}
 	remove(GETPCONFIG(s)->occtl_socket_file);
 	remove_pid_file();
 
+	snapshot_terminate(config_snapshot);
+
+	if (GETPCONFIG(s)->listen_netns_name && close_namespaces(&s->netns) < 0) {
+		fprintf(stderr, "cannot close listen namespaces\n");
+		exit(1);
+	}
 	clear_lists(s);
 	clear_vhosts(s->vconfig);
 	talloc_free(s->config_pool);
@@ -1457,3 +1625,119 @@ int main(int argc, char** argv)
 
 	return 0;
 }
+
+extern char ** pam_auth_group_list;
+extern char ** gssapi_auth_group_list;
+extern char ** plain_auth_group_list;
+extern unsigned pam_auth_group_list_size;
+extern unsigned gssapi_auth_group_list_size;
+extern unsigned plain_auth_group_list_size;
+
+static bool set_env_from_ws(main_server_st *s)
+{
+	worker_st *ws = s->ws;
+	WorkerStartupMsg msg = WORKER_STARTUP_MSG__INIT;
+	size_t msg_size;
+	uint8_t *msg_buffer = NULL;
+	size_t string_size = 0;
+	char *string_buffer = NULL;
+	int ret = 0;
+	SnapshotEntryMsg **entries = NULL;
+	SnapshotEntryMsg entry_template = SNAPSHOT_ENTRY_MSG__INIT;
+	size_t entry_count;
+	size_t index = 0;
+	struct htable_iter iter;
+
+	msg.secmod_addr.data = (uint8_t *)&ws->secmod_addr;
+	msg.secmod_addr.len = ws->secmod_addr_len;
+	msg.cmd_fd = ws->cmd_fd;
+	msg.conn_fd = ws->conn_fd;
+	msg.conn_type = (WorkerStartupMsg__CONNTYPE)ws->conn_type;
+	msg.remote_ip_str = ws->remote_ip_str;
+	msg.our_ip_str = ws->our_ip_str;
+	msg.session_start_time = ws->session_start_time;
+	msg.remote_addr.data = (uint8_t *)&ws->remote_addr;
+	msg.remote_addr.len = ws->remote_addr_len;
+	msg.our_addr.data = (uint8_t *)&ws->our_addr;
+	msg.our_addr.len = ws->our_addr_len;
+	msg.sec_auth_init_hmac.data = (uint8_t *)ws->sec_auth_init_hmac;
+	msg.sec_auth_init_hmac.len = sizeof(ws->sec_auth_init_hmac);
+
+	entry_count = snapshot_entry_count(config_snapshot);
+
+	entries = talloc_zero_array(s, SnapshotEntryMsg *, entry_count);
+	if (!entries)
+		goto cleanup;
+
+	for (index = 0; index < entry_count; index++) {
+		int fd, rr;
+		const char *file_name;
+		if (index == 0) {
+			rr = snapshot_first(config_snapshot, &iter, &fd, &file_name);
+		} else {
+			rr = snapshot_next(config_snapshot, &iter, &fd, &file_name);
+		}
+		if (rr < 0) {
+			mslog(s, NULL, LOG_ERR, "snapshot restoration failed (%d)\n", ret);
+			goto cleanup;
+		}
+
+		entries[index] = talloc_zero(s, SnapshotEntryMsg);
+		*entries[index] = entry_template;
+		entries[index]->file_descriptor = fd;
+		entries[index]->file_name = (char *)file_name;
+	}
+
+	msg.n_snapshot_entries = entry_count;
+	msg.snapshot_entries = entries;
+
+	msg.n_gssapi_auth_group_list = gssapi_auth_group_list_size;
+	msg.gssapi_auth_group_list = gssapi_auth_group_list;
+	msg.n_pam_auth_group_list = pam_auth_group_list_size;
+	msg.pam_auth_group_list = pam_auth_group_list;
+	msg.n_plain_auth_group_list = plain_auth_group_list_size;
+	msg.plain_auth_group_list = plain_auth_group_list;
+
+	msg_size = worker_startup_msg__get_packed_size(&msg);
+	if (msg_size == 0) {
+		mslog(s, NULL, LOG_ERR, "worker_startup_msg__get_packed_size failed\n");
+		goto cleanup;
+	}
+
+	msg_buffer = talloc_size(ws, msg_size);
+	if (!msg_buffer) {
+		mslog(s, NULL, LOG_ERR, "talloc_size failed\n");
+		goto cleanup;
+	}
+	msg_size = worker_startup_msg__pack(&msg, msg_buffer);
+	if (msg_size == 0) {
+		mslog(s, NULL, LOG_ERR, "worker_startup_msg__pack failed\n");
+		goto cleanup;
+	}
+	string_size = BASE64_ENCODE_RAW_LENGTH(msg_size) + 1;
+	string_buffer = talloc_size(ws, string_size);
+	if (!msg_buffer) {
+		mslog(s, NULL, LOG_ERR, "talloc_size failed\n");
+		goto cleanup;
+	}
+
+	oc_base64_encode((const char *)msg_buffer, msg_size, string_buffer, string_size);
+	if (setenv(OCSERV_ENV_WORKER_STARTUP_MSG, string_buffer, 1)) {
+		mslog(s, NULL, LOG_ERR, "setenv failed\n");
+		goto cleanup;
+	}
+
+	ret = 1;
+
+cleanup:
+	if (entries)
+		talloc_free(entries);
+
+	if (msg_buffer)
+		talloc_free(msg_buffer);
+
+	if (string_buffer)
+		talloc_free(string_buffer);
+
+	return ret;
+}
