@@ -6,7 +6,7 @@
  *
  * This file is part of ocserv.
  *
- * The GnuTLS is free software; you can redistribute it and/or
+ * ocserv is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License
  * as published by the Free Software Foundation; either version 2.1 of
  * the License, or (at your option) any later version.
@@ -24,6 +24,7 @@
 
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/resource.h>
 #include <unistd.h>
 #include <net/if.h>
 #include <vpn.h>
@@ -34,8 +35,9 @@
 #include <sys/uio.h>
 #include <signal.h>
 #include <ev.h>
-
+#include <hmac.h>
 #include "vhost.h"
+#include <namespace.h>
 
 #if defined(__FreeBSD__) || defined(__OpenBSD__)
 # include <limits.h>
@@ -52,7 +54,13 @@ extern ev_timer maintainance_watcher;
 
 #define MAIN_MAINTENANCE_TIME (900)
 
-int cmd_parser (void *pool, int argc, char **argv, struct list_head *head);
+int cmd_parser (void *pool, int argc, char **argv, struct list_head *head, bool worker);
+
+#if defined(CAPTURE_LATENCY_SUPPORT)
+#define LATENCY_AGGREGATION_TIME (60)
+#endif
+
+#define MINIMUM_USERS_PER_SEC_MOD 500
 
 struct listener_st {
 	ev_io io;
@@ -64,6 +72,7 @@ struct listener_st {
 	socklen_t addr_len;
 	int family;
 	int protocol;
+	ev_timer resume_accept;
 };
 
 struct listen_list_st {
@@ -116,6 +125,10 @@ typedef struct proc_st {
 	uint8_t sid[SID_SIZE];
 	unsigned active_sid;
 
+	/* non zero if the sid has been invalidated and must not be allowed
+	 * to reconnect. */
+	unsigned invalidated;
+
 	/* whether the host-update script has already been called */
 	unsigned host_updated;
 
@@ -133,11 +146,14 @@ typedef struct proc_st {
 	/* the following are copied here from the worker process for reporting
 	 * purposes (from main-ctl-handler). */
 	char user_agent[MAX_AGENT_NAME];
+	char device_type[MAX_DEVICE_TYPE];
+	char device_platform[MAX_DEVICE_PLATFORM];
 	char tls_ciphersuite[MAX_CIPHERSUITE_NAME];
 	char dtls_ciphersuite[MAX_CIPHERSUITE_NAME];
 	char cstp_compr[8];
 	char dtls_compr[8];
 	unsigned mtu;
+	unsigned int sec_mod_instance_index;
 
 	/* if the session is initiated by a cookie the following two are set
 	 * and are considered when generating an IP address. That is used to
@@ -190,6 +206,14 @@ struct proc_hash_db_st {
 	unsigned total;
 };
 
+#if defined(CAPTURE_LATENCY_SUPPORT)
+struct latency_stats_st {
+	uint64_t median_total;
+	uint64_t rms_total;
+	uint64_t sample_count;
+};
+#endif
+
 struct main_stats_st {
 	uint64_t session_timeouts; /* sessions with timeout */
 	uint64_t session_idle_timeouts; /* sessions with idle timeout */
@@ -201,15 +225,9 @@ struct main_stats_st {
 	unsigned max_mtu;
 
 	unsigned active_clients;
-	/* updated on the cli_stats_msg from sec-mod. 
-	 * Holds the number of entries in secmod list of users */
-	unsigned secmod_client_entries;
-	unsigned tlsdb_entries;
 	time_t start_time;
 	time_t last_reset;
 
-	uint32_t avg_auth_time; /* in seconds */
-	uint32_t max_auth_time; /* in seconds */
 	uint32_t avg_session_mins; /* in minutes */
 	uint32_t max_session_mins;
 	uint64_t auth_failures; /* authentication failures */
@@ -217,8 +235,38 @@ struct main_stats_st {
 	/* These are counted since start time */
 	uint64_t total_auth_failures; /* authentication failures since start_time */
 	uint64_t total_sessions_closed; /* sessions closed since start_time */
+
+#if defined(CAPTURE_LATENCY_SUPPORT)
+	struct latency_stats_st current_latency_stats;
+	struct latency_stats_st delta_latency_stats;
+#endif
 };
 
+typedef struct sec_mod_instance_st {
+	struct main_server_st * server;
+	char socket_file[_POSIX_PATH_MAX];
+	char full_socket_file[_POSIX_PATH_MAX];
+	pid_t sec_mod_pid;
+
+	struct sockaddr_un secmod_addr;
+	unsigned secmod_addr_len;
+
+	int sec_mod_fd; /* messages are sent and received async */
+	int sec_mod_fd_sync; /* messages are send in a sync order (ping-pong). Only main sends. */
+	/* updated on the cli_stats_msg from sec-mod. 
+	 * Holds the number of entries in secmod list of users */
+	unsigned secmod_client_entries;
+	unsigned tlsdb_entries;
+	uint32_t avg_auth_time; /* in seconds */
+	uint32_t max_auth_time; /* in seconds */
+
+} sec_mod_instance_st;
+
+typedef struct if_address_st {
+	struct sockaddr if_addr;
+	struct sockaddr if_netmask;
+} if_address_st;
+
 typedef struct main_server_st {
 	/* virtual hosts are only being added to that list, never removed */
 	struct list_head *vconfig;
@@ -232,13 +280,6 @@ typedef struct main_server_st {
 	struct script_list_st script_list;
 	/* maps DTLS session IDs to proc entries */
 	struct proc_hash_db_st proc_table;
-	
-	char socket_file[_POSIX_PATH_MAX];
-	char full_socket_file[_POSIX_PATH_MAX];
-	pid_t sec_mod_pid;
-
-	struct sockaddr_un secmod_addr;
-	unsigned secmod_addr_len;
 
 	struct main_stats_st stats;
 
@@ -247,22 +288,34 @@ typedef struct main_server_st {
 	/* This one is on worker pool */
 	struct worker_st *ws;
 
+	unsigned int sec_mod_instance_count;
+	sec_mod_instance_st * sec_mod_instances;
+
 	int top_fd;
 	int ctl_fd;
 
-	int sec_mod_fd; /* messages are sent and received async */
-	int sec_mod_fd_sync; /* messages are send in a sync order (ping-pong). Only main sends. */
 	void *main_pool; /* talloc main pool */
 	void *config_pool; /* talloc config pool */
 
+	const uint8_t hmac_key[HMAC_DIGEST_SIZE];
+
 	/* used as temporary buffer (currently by forward_udp_to_owner) */
 	uint8_t msg_buffer[MAX_MSG_SIZE];
+
+	struct netns_fds netns;
+
+#ifdef RLIMIT_NOFILE
+	struct rlimit fd_limits_default_set;
+#endif
+
+	struct if_address_st * if_addresses;
+	unsigned int if_addresses_count; 
 } main_server_st;
 
 void clear_lists(main_server_st *s);
 
 int handle_worker_commands(main_server_st *s, struct proc_st* cur);
-int handle_sec_mod_commands(main_server_st *s);
+int handle_sec_mod_commands(sec_mod_instance_st * sec_mod_instances);
 
 int user_connected(main_server_st *s, struct proc_st* cur);
 void user_hostname_update(main_server_st *s, struct proc_st* cur);
@@ -270,8 +323,8 @@ void user_disconnected(main_server_st *s, struct proc_st* cur);
 
 int send_udp_fd(main_server_st* s, struct proc_st * proc, int fd);
 
-int session_open(main_server_st * s, struct proc_st *proc, const uint8_t *cookie, unsigned cookie_size);
-int session_close(main_server_st * s, struct proc_st *proc);
+int session_open(sec_mod_instance_st * sec_mod_instance, struct proc_st *proc, const uint8_t *cookie, unsigned cookie_size);
+int session_close(sec_mod_instance_st * sec_mod_instance, struct proc_st *proc);
 
 #ifdef UNDER_TEST
 /* for testing */
@@ -306,13 +359,13 @@ int set_tun_mtu(main_server_st* s, struct proc_st * proc, unsigned mtu);
 int send_cookie_auth_reply(main_server_st* s, struct proc_st* proc,
 			AUTHREP r);
 
-int handle_auth_cookie_req(main_server_st* s, struct proc_st* proc,
+int handle_auth_cookie_req(sec_mod_instance_st * sec_mod_instance, struct proc_st* proc,
  			   const AuthCookieRequestMsg * req);
 
 int check_multiple_users(main_server_st *s, struct proc_st* proc);
 int handle_script_exit(main_server_st *s, struct proc_st* proc, int code);
 
-int run_sec_mod(main_server_st * s, int *sync_fd);
+void run_sec_mod(sec_mod_instance_st * sec_mod_instance, unsigned int instance_index);
 
 struct proc_st *new_proc(main_server_st * s, pid_t pid, int cmd_fd,
 			struct sockaddr_storage *remote_addr, socklen_t remote_addr_len,
@@ -327,14 +380,18 @@ struct proc_st *new_proc(main_server_st * s, pid_t pid, int cmd_fd,
 void remove_proc(main_server_st* s, struct proc_st *proc, unsigned flags);
 void proc_to_zombie(main_server_st* s, struct proc_st *proc);
 
-inline static void terminate_proc(main_server_st *s, proc_st *proc)
+inline static void disconnect_proc(main_server_st *s, proc_st *proc)
 {
-	/* if it has an IP, send a signal so that we cleanup
-	 * and get stats properly */
-	if (proc->pid != -1 && proc->pid != 0)
+	/* make sure that the SID cannot be reused */
+	proc->invalidated = 1;
+
+	/* if it has a PID, send a signal so that we cleanup
+	 * and sec-mod gets stats orderly */
+	if (proc->pid != -1 && proc->pid != 0) {
                 kill(proc->pid, SIGTERM);
-	else
+	} else {
 		remove_proc(s, proc, RPROC_KILL);
+	}
 }
 
 void put_into_cgroup(main_server_st * s, const char* cgroup, pid_t pid);
@@ -355,9 +412,10 @@ int send_socket_msg_to_worker(main_server_st* s, struct proc_st* proc, uint8_t c
 	return send_socket_msg(proc, proc->fd, cmd, socketfd, msg, get_size, pack);
 }
 
-int secmod_reload(main_server_st * s);
+int secmod_reload(sec_mod_instance_st * sec_mod_instance);
 
 const char *secmod_socket_file_name(struct perm_cfg_st *perm_config);
+void restore_secmod_socket_file_name(const char * save_path);
 void clear_vhosts(struct list_head *head);
 
 void request_reload(int signo);
