@@ -53,6 +53,7 @@
 #include <sec-mod-sup-config.h>
 #include <sec-mod-acct.h>
 #include <c-strcase.h>
+#include <hmac.h>
 
 #ifdef HAVE_GSSAPI
 # include <gssapi/gssapi.h>
@@ -301,7 +302,7 @@ int check_group(sec_mod_st * sec, client_entry_st * e)
 			}
 
 			if (found == 0) {
-				seclog(sec, LOG_AUTH, "user '%s' requested group '%s' but is not included on his certificate groups",
+				seclog(sec, LOG_NOTICE, "user '%s' requested group '%s' but is not included on his certificate groups",
 					e->acct_info.username, req_group);
 				return -1;
 			}
@@ -317,7 +318,7 @@ int check_group(sec_mod_st * sec, client_entry_st * e)
 	return 0;
 }
 
-/* Performs the required steps based on the result from the 
+/* Performs the required steps based on the result from the
  * authentication function (e.g. handle_auth_init).
  *
  * @cmd: the command received
@@ -339,7 +340,7 @@ int handle_sec_auth_res(int cfd, sec_mod_st * sec, client_entry_st * e, int resu
 			return ret;
 		}
 		e->msg_str = pst.msg_str;
-		
+
 		/* password requested for the next stage in multifactor auth OR password is requested again at the same stage */
 		passwd_retries = (e->passwd_counter < pst.counter) ? 0 : 1;
 		e->passwd_counter = pst.counter;
@@ -494,6 +495,9 @@ int handle_secm_session_open_cmd(sec_mod_st *sec, int fd, const SecmSessionOpenM
 
 	rep.username = e->acct_info.username;
 	rep.groupname = e->acct_info.groupname;
+	rep.user_agent = e->acct_info.user_agent;
+	rep.device_platform = e->acct_info.device_platform;
+	rep.device_type = e->acct_info.device_type;
 	rep.ip = e->acct_info.remote_ip;
 	rep.tls_auth_ok = e->tls_auth_ok;
 	rep.vhost = e->vhost->name;
@@ -558,7 +562,7 @@ int handle_secm_session_close_cmd(sec_mod_st *sec, int fd, const SecmSessionClos
 	e = find_client_entry(sec, req->sid.data);
 	if (e == NULL) {
 		seclog(sec, LOG_INFO, "session close but with non-existing SID");
-		return send_msg(e, fd, CMD_SECM_CLI_STATS, &rep,
+		return send_msg(sec, fd, CMD_SECM_CLI_STATS, &rep,
 		                (pack_size_func) cli_stats_msg__get_packed_size,
 		                (pack_func) cli_stats_msg__pack);
 	}
@@ -581,6 +585,10 @@ int handle_secm_session_close_cmd(sec_mod_st *sec, int fd, const SecmSessionClos
 			e->stats.bytes_out = req->bytes_out;
 	}
 
+	if (req->server_disconnected) {
+		e->discon_reason = REASON_SERVER_DISCONNECT;
+	}
+
 	/* send reply */
 	rep.bytes_in = e->stats.bytes_in;
 	rep.bytes_out = e->stats.bytes_out;
@@ -603,7 +611,6 @@ int handle_secm_session_close_cmd(sec_mod_st *sec, int fd, const SecmSessionClos
 	return 0;
 }
 
-
 void handle_sec_auth_ban_ip_reply(sec_mod_st *sec, const BanIpReplyMsg *msg)
 {
 	client_entry_st *e;
@@ -697,8 +704,8 @@ int handle_sec_auth_cont(int cfd, sec_mod_st * sec, const SecAuthContMsg * req)
 	}
 
 	if (e->status != PS_AUTH_INIT && e->status != PS_AUTH_CONT) {
-		seclog(sec, LOG_ERR, "auth cont received for %s "SESSION_STR" but we are on state %u!",
-		       e->acct_info.username, e->acct_info.safe_id, e->status);
+		seclog(sec, LOG_ERR, "auth cont received for %s "SESSION_STR" but we are on state %s(%u)!",
+		       e->acct_info.username, e->acct_info.safe_id, ps_status_to_str(e->status, 0), e->status);
 		ret = -1;
 		goto cleanup;
 	}
@@ -768,9 +775,41 @@ int handle_sec_auth_init(int cfd, sec_mod_st *sec, const SecAuthInitMsg *req, pi
 	unsigned i;
 	unsigned need_continue = 0;
 	vhost_cfg_st *vhost;
+	hmac_component_st hmac_components[3];
+	uint8_t computed_hmac[HMAC_DIGEST_SIZE];
+	time_t now = time(0);
+	time_t session_start_time;
+
+	if (req->hmac.len != HMAC_DIGEST_SIZE || !req->hmac.data) {
+		seclog(sec, LOG_NOTICE, "hmac is the wrong size");
+		return -1;
+	}
+
+	/* Authenticate the client parameters */
+	session_start_time = (time_t)req->session_start_time; // avoid time_t size problem
+
+	hmac_components[0].data =  req->ip;
+	// req->ip is required and protobuf doesn't permit null for required parameters
+	hmac_components[0].length = strlen(req->ip);
+	hmac_components[1].data = req->our_ip;
+	hmac_components[1].length = req->our_ip ? strlen(req->our_ip) : 0;
+	hmac_components[2].data = (void*)&session_start_time;
+	hmac_components[2].length = sizeof(session_start_time);
+
+	generate_hmac(sizeof(sec->hmac_key), sec->hmac_key, sizeof(hmac_components) / sizeof(hmac_components[0]), hmac_components, computed_hmac);
+
+	if (memcmp(computed_hmac, req->hmac.data, req->hmac.len) != 0) {
+		seclog(sec, LOG_NOTICE, "hmac presented by client doesn't match parameters provided - possible replay");
+		return -1;
+	}
 
 	vhost = find_vhost(sec->vconfig, req->vhost);
 
+	if ((now - session_start_time) > vhost->perm_config.config->auth_timeout) {
+		seclog(sec, LOG_NOTICE, "hmac presented by client expired - possible replay");
+		return -1;
+	}
+
 	e = new_client_entry(sec, vhost, req->ip, pid);
 	if (e == NULL) {
 		seclog(sec, LOG_ERR, "cannot initialize memory");
@@ -803,11 +842,22 @@ int handle_sec_auth_init(int cfd, sec_mod_st *sec, const SecAuthInitMsg *req, pi
 
 	e->tls_auth_ok = req->tls_auth_ok;
 
+	if (req->device_platform != NULL) {
+		strlcpy(e->acct_info.device_platform, req->device_platform, sizeof(e->acct_info.device_platform));
+	}
+
+	if (req->device_type != NULL) {
+		strlcpy(e->acct_info.device_type, req->device_type, sizeof(e->acct_info.device_type));
+	}
+
 	if (req->user_agent != NULL)
 		strlcpy(e->acct_info.user_agent, req->user_agent, sizeof(e->acct_info.user_agent));
 
-	if (req->user_name != NULL) {
-		strlcpy(e->acct_info.username, req->user_name, sizeof(e->acct_info.username));
+	// Real user name is retrieved after auth.
+	if (!(req->auth_type & CONFIDENTIAL_USER_NAME_AUTH_TYPES)) {
+		if (req->user_name != NULL) {
+			strlcpy(e->acct_info.username, req->user_name, sizeof(e->acct_info.username));
+		}
 	}
 
 	if (req->our_ip != NULL) {
@@ -832,7 +882,7 @@ int handle_sec_auth_init(int cfd, sec_mod_st *sec, const SecAuthInitMsg *req, pi
 	}
 
 	e->status = PS_AUTH_INIT;
-	seclog(sec, LOG_DEBUG, "auth init %sfor user '%s' "SESSION_STR" of group: '%s' from '%s'", 
+	seclog(sec, LOG_DEBUG, "auth init %sfor user '%s' "SESSION_STR" of group: '%s' from '%s'",
 	       req->tls_auth_ok?"(with cert) ":"",
 	       e->acct_info.username, e->acct_info.safe_id, e->acct_info.groupname, req->ip);
 
