@@ -58,12 +58,15 @@ void cstp_cork(worker_st *ws)
 	if (ws->session) {
 		gnutls_record_cork(ws->session);
 	} else {
-		int state = 1;
-#ifdef __linux__
-		setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_CORK, &state, sizeof(state));
+		int state = 1, ret = 0;
+#if defined(__linux__)
+		ret = setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_CORK, &state, sizeof(state));
 #elif defined(TCP_NOPUSH)
-		setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_NOPUSH, &state, sizeof(state));
+		ret = setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_NOPUSH, &state, sizeof(state));
 #endif
+		if (ret == -1) {
+			oclog(ws, LOG_ERR, "setsockopt(IPPROTO_TCP(TCP_CORK) failed");
+		}
 	}
 }
 
@@ -72,12 +75,15 @@ int cstp_uncork(worker_st *ws)
 	if (ws->session) {
 		return gnutls_record_uncork(ws->session, GNUTLS_RECORD_WAIT);
 	} else {
-		int state = 0;
+		int state = 0, ret = 0;
 #if defined(__linux__)
-		setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_CORK, &state, sizeof(state));
+		ret = setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_CORK, &state, sizeof(state));
 #elif defined(TCP_NOPUSH)
-		setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_NOPUSH, &state, sizeof(state));
+		ret = setsockopt(ws->conn_fd, IPPROTO_TCP, TCP_NOPUSH, &state, sizeof(state));
 #endif
+		if (ret == -1) {
+			oclog(ws, LOG_ERR, "setsockopt(IPPROTO_TCP(TCP_UNCORK) failed");
+		}
 		return 0;
 	}
 }
@@ -115,17 +121,25 @@ ssize_t cstp_send(worker_st *ws, const void *data,
 
 ssize_t cstp_send_file(worker_st *ws, const char *file)
 {
-int fd;
-char buf[512];
-ssize_t len, total = 0;
-int ret;
+	int fd;
+	char buf[1024];
+	int counter = 100; /* allow 10 seconds for a full packet */
+	ssize_t len, total = 0;
+	int ret;
 
 	fd = open(file, O_RDONLY);
 	if (fd == -1)
 		return GNUTLS_E_FILE_ERROR;
 
 	while (	(len = read( fd, buf, sizeof(buf))) > 0 ||
-		(len == -1 && (errno == EINTR || errno == EAGAIN))) {
+		(len == -1 && counter > 0 && (errno == EINTR || errno == EAGAIN))) {
+
+		if (len == -1) {
+			counter--;
+			ms_sleep(100);
+			continue;
+		}
+
 		ret = cstp_send(ws, buf, len);
 		CSTP_FATAL_ERR(ws, ret);
 
@@ -325,13 +339,13 @@ void cstp_fatal_close(worker_st *ws,
 	}
 }
 
-ssize_t dtls_recv_packet(worker_st *ws, gnutls_datum_t *data, void **p)
+ssize_t dtls_recv_packet(struct dtls_st *dtls, gnutls_datum_t *data, void **p)
 {
 	int ret;
 #ifdef ZERO_COPY
 	gnutls_packet_t packet = NULL;
 
-	ret = gnutls_record_recv_packet(ws->dtls_session, &packet);
+	ret = gnutls_record_recv_packet(dtls->dtls_session, &packet);
 	if (ret > 0) {
 		gnutls_packet_get(packet, data, NULL);
 		*p = packet;
@@ -340,7 +354,7 @@ ssize_t dtls_recv_packet(worker_st *ws, gnutls_datum_t *data, void **p)
 	}
 #else
 	ret =
-	    gnutls_record_recv(ws->dtls_session, ws->buffer, ws->buffer_size);
+	    gnutls_record_recv(dtls->dtls_session, ws->buffer, ws->buffer_size);
 	data->data = ws->buffer;
 	data->size = ret;
 #endif
@@ -348,7 +362,7 @@ ssize_t dtls_recv_packet(worker_st *ws, gnutls_datum_t *data, void **p)
 	return ret;
 }
 
-ssize_t dtls_send(worker_st *ws, const void *data,
+ssize_t dtls_send(struct dtls_st *dtls, const void *data,
 			size_t data_size)
 {
 	int ret;
@@ -356,7 +370,7 @@ ssize_t dtls_send(worker_st *ws, const void *data,
 	const uint8_t* p = data;
 
 	while(left > 0) {
-		ret = gnutls_record_send(ws->dtls_session, p, data_size);
+		ret = gnutls_record_send(dtls->dtls_session, p, data_size);
 		if (ret < 0) {
 			if (ret != GNUTLS_E_AGAIN && ret != GNUTLS_E_INTERRUPTED) {
 				return ret;
@@ -375,10 +389,10 @@ ssize_t dtls_send(worker_st *ws, const void *data,
 	return data_size;
 }
 
-void dtls_close(worker_st *ws)
+void dtls_close(struct dtls_st *dtls)
 {
-	gnutls_bye(ws->dtls_session, GNUTLS_SHUT_WR);
-	gnutls_deinit(ws->dtls_session);
+	gnutls_bye(dtls->dtls_session, GNUTLS_SHUT_WR);
+	gnutls_deinit(dtls->dtls_session);
 }
 
 static size_t rehash(const void *_e, void *unused)
@@ -430,8 +444,10 @@ static void tls_audit_log_func(gnutls_session_t session, const char *str)
 {
 	worker_st * ws;
 
+	(void)(ws);
+
 	if (session == NULL)
-		syslog(LOG_AUTH, "warning: %s", str);
+		syslog(LOG_NOTICE, "warning: %s", str);
 	else {
 		ws = gnutls_session_get_ptr(session);
 
@@ -451,7 +467,8 @@ static int verify_certificate_cb(gnutls_session_t session)
 		return -1;
 	}
 
-	if (session == ws->dtls_session) {
+	if ((session == DTLS_ACTIVE(ws)->dtls_session) ||
+		(session == DTLS_INACTIVE(ws)->dtls_session)) {
 		oclog(ws, LOG_ERR, "unexpected issue; client shouldn't have offered a certificate in DTLS");
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
@@ -568,6 +585,8 @@ static void certificate_check(main_server_st *s, const char *vhostname, gnutls_p
 	const char *cert_name = "unnamed";
 	time_t t;
 
+	(void)cert_name;
+
 	ret = gnutls_x509_crt_init(&crt);
 	GNUTLS_FATAL_ERR(ret);
 
