@@ -118,6 +118,21 @@ int os_set_ipv6_addr(main_server_st * s, struct proc_st *proc)
 		goto cleanup;
 	}
 
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_addr.sa_family = AF_INET6;
+	ifr.ifr_flags |= IFF_UP | IFF_RUNNING;
+	strlcpy(ifr.ifr_name, proc->tun_lease.name, IFNAMSIZ);
+
+	ret = ioctl(fd, SIOCSIFFLAGS, &ifr);
+	if (ret != 0) {
+		e = errno;
+		mslog(s, NULL, LOG_ERR,
+		      "%s: Could not bring up IPv6 interface: %s\n",
+		      proc->tun_lease.name, strerror(e));
+		ret = -1;
+		goto cleanup;
+	}
+
 	/* route to our remote address */
 	memset(&rt6, 0, sizeof(rt6));
 	memcpy(&rt6.rtmsg_dst, SA_IN6_P(&proc->ipv6->rip),
@@ -138,21 +153,6 @@ int os_set_ipv6_addr(main_server_st * s, struct proc_st *proc)
 		goto cleanup;
 	}
 
-	memset(&ifr, 0, sizeof(ifr));
-	ifr.ifr_addr.sa_family = AF_INET6;
-	ifr.ifr_flags |= IFF_UP | IFF_RUNNING;
-	strlcpy(ifr.ifr_name, proc->tun_lease.name, IFNAMSIZ);
-
-	ret = ioctl(fd, SIOCSIFFLAGS, &ifr);
-	if (ret != 0) {
-		e = errno;
-		mslog(s, NULL, LOG_ERR,
-		      "%s: Could not bring up IPv6 interface: %s\n",
-		      proc->tun_lease.name, strerror(e));
-		ret = -1;
-		goto cleanup;
-	}
-
 	ret = 0;
  cleanup:
 	close(fd);
@@ -306,7 +306,7 @@ static void os_reset_ipv6_addr(struct proc_st *proc)
 		ifr6.ifr_addr.sin6_len = sizeof(struct sockaddr_in6);
 		ifr6.ifr_addr.sin6_family = AF_INET6;
 
-		ioctl(fd, SIOCDIFADDR_IN6, &ifr6);
+		(void)ioctl(fd, SIOCDIFADDR_IN6, &ifr6);
 		close(fd);
 	}
 }
@@ -802,7 +802,7 @@ static void reset_ipv4_addr(struct proc_st *proc)
 		ifr.ifr_addr.sa_len = sizeof(struct sockaddr_in);
 		ifr.ifr_addr.sa_family = AF_INET;
 
-		ioctl(fd, SIOCDIFADDR, &ifr);
+		(void)ioctl(fd, SIOCDIFADDR, &ifr);
 		close(fd);
 	}
 #elif defined(__linux__)
@@ -814,7 +814,7 @@ static void reset_ipv4_addr(struct proc_st *proc)
 		memcpy(&ifr.ifr_addr, &proc->ipv4->lip, proc->ipv4->lip_len);
 		ifr.ifr_addr.sa_family = AF_INET;
 
-		ioctl(fd, SIOCDIFADDR, &ifr);
+		(void)ioctl(fd, SIOCDIFADDR, &ifr);
 
 		close(fd);
 	}
