@@ -73,41 +73,49 @@ static const char ocv3_success_msg_head[] = "<?xml version=\"1.0\" encoding=\"UT
 
 static const char ocv3_success_msg_foot[] = "</auth>\n";
 
-#define OC_LOGIN_MSG_START \
+#define OC_LOGIN_START \
     "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" \
     "<config-auth client=\"vpn\" type=\"auth-request\">\n" \
     VERSION_MSG \
-    "<auth id=\"main\">\n" \
+    "<auth id=\"main\">\n"
+
+#define OC_LOGIN_FORM_START \
     "<message>%s</message>\n" \
     "<form method=\"post\" action=\"/auth\">\n"
 
-static const char oc_login_msg_end[] =
-    "</form></auth>\n" "</config-auth>";
+#define OC_LOGIN_END \
+    "</form></auth>\n" "</config-auth>"
 
-static const char login_msg_user[] =
-    "<input type=\"text\" name=\"username\" label=\"Username:\" />\n";
+#define OC_LOGIN_FORM_INPUT_USER \
+    "<input type=\"text\" name=\"username\" label=\"Username:\" />\n"
 
 #define DEFAULT_PASSWD_LABEL "Password:"
-#define LOGIN_MSG_PASSWORD \
+#define OC_LOGIN_FORM_INPUT_PASSWORD \
     "<input type=\"password\" name=\"password\" label=\""DEFAULT_PASSWD_LABEL"\" />\n"
-#define LOGIN_MSG_PASSWORD_CTR \
+#define OC_LOGIN_FORM_INPUT_PASSWORD_CTR \
     "<input type=\"password\" name=\"secondary_password\" label=\"Password%d:\" />\n"
 
 #define _OCV3_LOGIN_MSG_START(x) \
     "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" \
-    "<auth id=\""x"\">\n" \
-    "<message>%s</message>\n" \
-    "<form method=\"post\" action=\"/auth\">\n"
+    "<auth id=\""x"\">\n"
 
-#define OCV3_LOGIN_MSG_START _OCV3_LOGIN_MSG_START("main")
-#define OCV3_PASSWD_MSG_START _OCV3_LOGIN_MSG_START("passwd")
+#define OCV3_LOGIN_START _OCV3_LOGIN_MSG_START("main")
+#define OCV3_PASSWD_START _OCV3_LOGIN_MSG_START("passwd")
+
+#define OCV3_LOGIN_END "</form></auth>\n"
+
+#ifdef SUPPORT_OIDC_AUTH
+#define HTTP_AUTH_OIDC_PREFIX "Bearer"
+#endif
 
-static const char ocv3_login_msg_end[] =
-    "</form></auth>\n";
 
 static int get_cert_info(worker_st * ws);
 static int basic_auth_handler(worker_st * ws, unsigned http_ver, const char *msg);
 
+#ifdef SUPPORT_OIDC_AUTH
+static int oidc_auth_handler(worker_st * ws, unsigned http_ver);
+#endif
+
 int ws_switch_auth_to(struct worker_st *ws, unsigned auth)
 {
 	unsigned i;
@@ -200,20 +208,20 @@ int get_auth_handler2(worker_st * ws, unsigned http_ver, const char *pmsg, unsig
 	char context[BASE64_ENCODE_RAW_LENGTH(SID_SIZE) + 1];
 	unsigned int i, j;
 	str_st str;
-	const char *login_msg_start;
-	const char *login_msg_end;
+	const char *login_start;
+	const char *login_end;
 
 	if (ws->req.user_agent_type == AGENT_OPENCONNECT_V3) {
 		/* certain v2.x modified clients require a different auth_id
 		 * when password is being requested, rather than username */
 		if (ws->auth_state == S_AUTH_REQ)
-			login_msg_start = OCV3_PASSWD_MSG_START;
+			login_start = OCV3_PASSWD_START;
 		else
-			login_msg_start = OCV3_LOGIN_MSG_START;
-		login_msg_end = ocv3_login_msg_end;
+			login_start = OCV3_LOGIN_START;
+		login_end = OCV3_LOGIN_END;
 	} else {
-		login_msg_start = OC_LOGIN_MSG_START;
-		login_msg_end = oc_login_msg_end;
+		login_start = OC_LOGIN_START;
+		login_end = OC_LOGIN_END;
 	}
 
 	if ((ws->selected_auth->type & AUTH_TYPE_GSSAPI) && ws->auth_state < S_AUTH_COOKIE) {
@@ -261,44 +269,64 @@ int get_auth_handler2(worker_st * ws, unsigned http_ver, const char *pmsg, unsig
 	}
 
 	if (ws->auth_state == S_AUTH_REQ) {
-		/* only ask password */
+		/* Password Form */
 		if (pmsg == NULL || strncasecmp(pmsg, DEFAULT_PASSWD_LABEL, sizeof(DEFAULT_PASSWD_LABEL)-1) == 0)
 			pmsg = "Please enter your password.";
 
-		ret = str_append_printf(&str, login_msg_start, pmsg);
+		ret = str_append_str(&str, login_start);
+		if (ret < 0) {
+			ret = -1;
+			goto cleanup;
+		}
+
+		ret = str_append_printf(&str, OC_LOGIN_FORM_START, pmsg);
 		if (ret < 0) {
 			ret = -1;
 			goto cleanup;
 		}
 
 		if (pcounter > 0)
-			ret = str_append_printf(&str, LOGIN_MSG_PASSWORD_CTR, pcounter);
+			ret = str_append_printf(&str, OC_LOGIN_FORM_INPUT_PASSWORD_CTR, pcounter);
 		else
-			ret = str_append_str(&str, LOGIN_MSG_PASSWORD);
+			ret = str_append_str(&str, OC_LOGIN_FORM_INPUT_PASSWORD);
 		if (ret < 0) {
 			ret = -1;
 			goto cleanup;
 		}
 
-		ret = str_append_str(&str, login_msg_end);
+		ret = str_append_str(&str, login_end);
 		if (ret < 0) {
 			ret = -1;
 			goto cleanup;
 		}
 
 	} else {
+		/* Username / Groups Form */
 		if (pmsg == NULL)
 			pmsg = "Please enter your username.";
 
-		/* ask for username and groups */
-		ret = str_append_printf(&str, login_msg_start, pmsg);
+		ret = str_append_str(&str, login_start);
+		if (ret < 0) {
+			ret = -1;
+			goto cleanup;
+		}
+
+		if (WSCONFIG(ws)->pre_login_banner) {
+			ret = str_append_printf(&str, "<banner>%s</banner>", WSCONFIG(ws)->pre_login_banner);
+			if (ret < 0) {
+				ret = -1;
+				goto cleanup;
+			}
+		}
+
+		ret = str_append_printf(&str, OC_LOGIN_FORM_START, pmsg);
 		if (ret < 0) {
 			ret = -1;
 			goto cleanup;
 		}
 
 		if (ws->selected_auth->type & AUTH_TYPE_USERNAME_PASS) {
-			ret = str_append_str(&str, login_msg_user);
+			ret = str_append_str(&str, OC_LOGIN_FORM_INPUT_USER);
 			if (ret < 0) {
 				ret = -1;
 				goto cleanup;
@@ -388,7 +416,7 @@ int get_auth_handler2(worker_st * ws, unsigned http_ver, const char *pmsg, unsig
 			}
 		}
 
-		ret = str_append_str(&str, login_msg_end);
+		ret = str_append_str(&str, login_end);
 		if (ret < 0) {
 			ret = -1;
 			goto cleanup;
@@ -447,6 +475,7 @@ int get_cert_names(worker_st * ws, const gnutls_datum_t * raw)
 	int ret;
 	unsigned i;
 	size_t size;
+	char cert_username[MAX_USERNAME_SIZE];
 
 	if (ws->cert_username[0] != 0 || ws->cert_groups_size > 0)
 		return 0; /* already read, nothing to do */
@@ -470,29 +499,37 @@ int get_cert_names(worker_st * ws, const gnutls_datum_t * raw)
 			size = sizeof(ws->cert_username);
 			ret =
 			    gnutls_x509_crt_get_subject_alt_name(crt, i,
-								 ws->
 								 cert_username,
 								 &size, NULL);
-			if (ret < 0)
+			if (ret < 0) {
+				if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
+					ret = 1;
 				break;
+			}
+
 			if (ret == GNUTLS_SAN_RFC822NAME) {
+				strlcpy(ws->cert_username, cert_username, sizeof(ws->cert_username));
 				oclog(ws, LOG_INFO,
 				      "RFC822NAME (%s) retrieved",
-				      ws->cert_username);
+				      cert_username);
 				break;
 			}
 		}
-		if (ret != 0) {
-			ret = 1;
-		}
-	} else if (WSCONFIG(ws)->cert_user_oid) {	/* otherwise certificate username is ignored */
+
+	} else if (WSCONFIG(ws)->cert_user_oid) { /* otherwise we check at the DN */
 		size = sizeof(ws->cert_username);
 		ret =
 		    gnutls_x509_crt_get_dn_by_oid(crt,
 					  WSCONFIG(ws)->cert_user_oid, 0,
-					  0, ws->cert_username, &size);
+					  0, cert_username, &size);
+		if (ret >= 0)
+			strlcpy(ws->cert_username, cert_username, sizeof(ws->cert_username));
+
 	} else {
-		ret = gnutls_x509_crt_get_dn(crt, ws->cert_username, &size);
+		size = sizeof(ws->cert_username);
+		ret = gnutls_x509_crt_get_dn(crt, cert_username, &size);
+		if (ret >= 0)
+			strlcpy(ws->cert_username, cert_username, sizeof(ws->cert_username));
 	}
 
 	if (ret < 0) {
@@ -831,7 +868,7 @@ int get_cert_info(worker_st * ws)
 			return -1;
 	}
 
-	/* this is superflous. Verification has already been performed 
+	/* this is superflous. Verification has already been performed
 	 * during handshake. */
 	cert = gnutls_certificate_get_peers(ws->session, &ncerts);
 
@@ -915,6 +952,8 @@ int auth_cookie(worker_st * ws, void *cookie, size_t cookie_size)
 
 	msg.cookie.data = cookie;
 	msg.cookie.len = cookie_size;
+	if (ws->req.hostname[0] != 0)
+		msg.hostname = ws->req.hostname;
 
 	ret = send_msg_to_main(ws, AUTH_COOKIE_REQ, &msg, (pack_size_func)
 			       auth_cookie_request_msg__get_packed_size,
@@ -953,10 +992,15 @@ int post_common_handler(worker_st * ws, unsigned http_ver, const char *imsg)
 		success_msg_foot_size = strlen(success_msg_foot);
 	} else {
 		success_msg_head = oc_success_msg_head;
+		success_msg_foot = NULL;
+#ifdef ANYCONNECT_CLIENT_COMPAT
 		if (WSCONFIG(ws)->xml_config_file) {
 			success_msg_foot = talloc_asprintf(ws, OC_SUCCESS_MSG_FOOT_PROFILE,
 				WSCONFIG(ws)->xml_config_file, WSCONFIG(ws)->xml_config_hash);
-		} else {
+		} 
+#endif
+
+		if (success_msg_foot == NULL) {
 			success_msg_foot = talloc_strdup(ws, OC_SUCCESS_MSG_FOOT);
 		}
 
@@ -1045,6 +1089,7 @@ int post_common_handler(worker_st * ws, unsigned http_ver, const char *imsg)
 	if (ret < 0)
 		goto fail;
 
+#ifdef ANYCONNECT_CLIENT_COMPAT	
 	if (WSCONFIG(ws)->xml_config_file) {
 		ret =
 		    cstp_printf(ws,
@@ -1058,6 +1103,7 @@ int post_common_handler(worker_st * ws, unsigned http_ver, const char *imsg)
 			       "Set-Cookie: webvpnc=bu:/&p:t&iu:1/&sh:%s; path=/; Secure\r\n",
 			       WSPCONFIG(ws)->cert_hash);
 	}
+#endif
 
 	if (ret < 0)
 		goto fail;
@@ -1330,26 +1376,48 @@ int basic_auth_handler(worker_st * ws, unsigned http_ver, const char *msg)
 	return ret;
 }
 
-static char *get_our_ip(worker_st * ws, char str[MAX_IP_STR])
+#ifdef SUPPORT_OIDC_AUTH
+static
+int oidc_auth_handler(worker_st * ws, unsigned http_ver)
 {
 	int ret;
-	struct sockaddr_storage sockaddr;
-	socklen_t socklen;
 
-	if (ws->our_addr_len > 0) {
-		return human_addr2((struct sockaddr*)&ws->our_addr, ws->our_addr_len, str, MAX_IP_STR, 0);
+	oclog(ws, LOG_HTTP_DEBUG, "HTTP sending: 401 Unauthorized");
+	cstp_cork(ws);
+	ret = cstp_printf(ws, "HTTP/1.%u 401 Unauthorized\r\n", http_ver);
+	if (ret < 0)
+		return -1;
+
+	oclog(ws, LOG_HTTP_DEBUG, "HTTP sending: WWW-Authenticate: %s", HTTP_AUTH_OIDC_PREFIX);
+	ret = cstp_printf(ws, "WWW-Authenticate: %s\r\n", HTTP_AUTH_OIDC_PREFIX);
+
+	if (ret < 0)
+		return -1;
+
+	ret = cstp_puts(ws, "Content-Length: 0\r\n");
+	if (ret < 0) {
+		ret = -1;
+		goto cleanup;
 	}
 
-	if (ws->udp_state != UP_ACTIVE)
-		return NULL;
+	ret = cstp_puts(ws, "\r\n");
+	if (ret < 0) {
+		ret = -1;
+		goto cleanup;
+	}
 
-	socklen = sizeof(sockaddr);
-	ret = getsockname(ws->dtls_tptr.fd, (struct sockaddr*)&sockaddr, &socklen);
-	if (ret == -1)
-		return NULL;
+	ret = cstp_uncork(ws);
+	if (ret < 0) {
+		ret = -1;
+		goto cleanup;
+	}
 
-	return human_addr2((struct sockaddr*)&sockaddr, socklen, str, MAX_IP_STR, 0);
+	ret = 0;
+
+ cleanup:
+	return ret;
 }
+#endif
 
 #define USERNAME_FIELD "username"
 #define GROUPNAME_FIELD "group%5flist"
@@ -1369,7 +1437,6 @@ int post_auth_handler(worker_st * ws, unsigned http_ver)
 	char *username = NULL;
 	char *password = NULL;
 	char *groupname = NULL;
-	char our_ip_str[MAX_IP_STR];
 	char *msg = NULL;
 	unsigned def_group = 0;
 	unsigned pcounter = 0;
@@ -1403,12 +1470,50 @@ int post_auth_handler(worker_st * ws, unsigned http_ver)
 				   strcmp(groupname, WSCONFIG(ws)->default_select_group) == 0) {
 				def_group = 1;
 			} else {
-				strlcpy(ws->groupname, groupname, sizeof(ws->groupname));
+				/* Some anyconnect clients send the group friendly name instead of
+				 * the actual value; see #267 */
+				ws->groupname[0] = 0;
+				if (WSCONFIG(ws)->friendly_group_list != NULL) {
+					unsigned found = 0, i;
+
+					for (i=0;i<WSCONFIG(ws)->group_list_size;i++) {
+						if (strcmp(WSCONFIG(ws)->group_list[i], groupname) == 0) {
+							found = 1;
+							break;
+						}
+					}
+
+					if (!found)
+						for (i=0;i<WSCONFIG(ws)->group_list_size;i++) {
+							if (WSCONFIG(ws)->friendly_group_list[i] != NULL && strcmp(WSCONFIG(ws)->friendly_group_list[i], groupname) == 0) {
+								strlcpy(ws->groupname, WSCONFIG(ws)->group_list[i], sizeof(ws->groupname));
+								break;
+							}
+						}
+				}
+
+				if (ws->groupname[0] == 0)
+					strlcpy(ws->groupname, groupname, sizeof(ws->groupname));
 				ireq.group_name = ws->groupname;
 			}
 		}
 		talloc_free(groupname);
 
+#ifdef SUPPORT_OIDC_AUTH
+		if (ws->selected_auth->type & AUTH_TYPE_OIDC) {
+			if (req->authorization == NULL || req->authorization_size == 0)
+				return oidc_auth_handler(ws, http_ver);
+
+			if ((req->authorization_size > (sizeof(HTTP_AUTH_OIDC_PREFIX) - 1)) && strncasecmp(req->authorization, HTTP_AUTH_OIDC_PREFIX, sizeof(HTTP_AUTH_OIDC_PREFIX) - 1) == 0) {
+				ireq.auth_type |= AUTH_TYPE_OIDC;
+				ireq.user_name = req->authorization + sizeof(HTTP_AUTH_OIDC_PREFIX);
+			} else {
+				oclog(ws, LOG_HTTP_DEBUG, "Invalid authorization data: %.*s", req->authorization_size, req->authorization);
+				goto auth_fail;
+			}
+		}
+#endif
+
 		if (ws->selected_auth->type & AUTH_TYPE_GSSAPI) {
 			if (req->authorization == NULL || req->authorization_size == 0)
 				return basic_auth_handler(ws, http_ver, NULL);
@@ -1475,10 +1580,19 @@ int post_auth_handler(worker_st * ws, unsigned http_ver)
 
 		ireq.vhost = ws->vhost->name;
 		ireq.ip = ws->remote_ip_str;
-		ireq.our_ip = get_our_ip(ws, our_ip_str);
+		ireq.our_ip = ws->our_ip_str;
+		ireq.session_start_time = ws->session_start_time;
+		ireq.hmac.data = (uint8_t*)ws->sec_auth_init_hmac;
+		ireq.hmac.len = sizeof(ws->sec_auth_init_hmac);
 		if (req->user_agent[0] != 0)
 			ireq.user_agent = req->user_agent;
 
+		if (req->devtype[0] != 0)
+			ireq.device_type = req->devtype;
+
+		if (req->devplatform[0] != 0)
+			ireq.device_platform = req->devplatform;
+
 		sd = connect_to_secmod(ws);
 		if (sd == -1) {
 			reason = MSG_INTERNAL_ERROR;
@@ -1503,7 +1617,6 @@ int post_auth_handler(worker_st * ws, unsigned http_ver)
 		SecAuthContMsg areq = SEC_AUTH_CONT_MSG__INIT;
 
 		areq.ip = ws->remote_ip_str;
-
 		if (ws->selected_auth->type & AUTH_TYPE_GSSAPI) {
 			if (req->authorization == NULL || req->authorization_size <= 10) {
 				if (req->authorization != NULL)
@@ -1574,6 +1687,7 @@ int post_auth_handler(worker_st * ws, unsigned http_ver)
 	}
 
 	if (ret == ERR_AUTH_CONTINUE) {
+		
 		oclog(ws, LOG_DEBUG, "continuing authentication for '%s'",
 		      ws->username);
 		ws->auth_state = S_AUTH_REQ;
@@ -1616,13 +1730,15 @@ int post_auth_handler(worker_st * ws, unsigned http_ver)
 	if (sd != -1)
 		close(sd);
 	oclog(ws, LOG_HTTP_DEBUG, "HTTP sending: 401 Unauthorized");
-	cstp_printf(ws,
+	ret = cstp_printf(ws,
 		   "HTTP/1.%d 401 %s\r\nContent-Length: 0\r\n\r\n",
 		   http_ver, reason);
-	cstp_fatal_close(ws, GNUTLS_A_ACCESS_DENIED);
+	if (ret >= 0)
+		cstp_fatal_close(ws, GNUTLS_A_ACCESS_DENIED);
 	talloc_free(msg);
 	exit_worker(ws);
  cleanup:
  	talloc_free(msg);
  	return ret;
 }
+
