@@ -60,8 +60,8 @@ static const struct known_urls_st known_urls[] = {
 	LL("/cert.cer", get_cert_der_handler, NULL),
 	LL("/ca.pem", get_ca_handler, NULL),
 	LL("/ca.cer", get_ca_der_handler, NULL),
-	LL_DIR("/profiles", get_config_handler, NULL),
 #ifdef ANYCONNECT_CLIENT_COMPAT
+	LL_DIR("/profiles", get_config_handler, NULL),
 	LL("/1/index.html", get_empty_handler, NULL),
 	LL("/1/Linux", get_empty_handler, NULL),
 	LL("/1/Linux_64", get_empty_handler, NULL),
@@ -171,6 +171,14 @@ static const dtls_ciphersuite_st ciphersuites12[] = {
 	}
 };
 
+#define STR_ST(x) {.data = (uint8_t*)x, .length = sizeof(x)-1}
+static const str_st sensitve_http_headers[] = {
+	STR_ST("Cookie"),
+	STR_ST("X-DTLS-Master-Secret"),
+	STR_ST("Authorization"),
+	{NULL, 0}
+};
+
 #ifdef HAVE_LZ4
 /* Wrappers over LZ4 functions */
 static
@@ -207,7 +215,6 @@ struct compression_method_st comp_methods[] = {
 		.server_prio = 80,
 	}
 };
-#endif
 
 unsigned switch_comp_priority(void *pool, const char *modstring)
 {
@@ -249,6 +256,18 @@ unsigned switch_comp_priority(void *pool, const char *modstring)
 	talloc_free(str);
 	return ret;
 }
+#endif
+
+static bool header_is_sensitive(str_st * header)
+{
+	size_t i;
+	for (i = 0; sensitve_http_headers[i].length != 0; i++) {
+		if ((header->length == sensitve_http_headers[i].length) &&
+			(strncasecmp((char*)header->data, (char*)sensitve_http_headers[i].data, header->length) == 0))
+			return true;
+	}
+	return false;
+}
 
 static
 void header_value_check(struct worker_st *ws, struct http_req_st *req)
@@ -268,9 +287,7 @@ void header_value_check(struct worker_st *ws, struct http_req_st *req)
 	if (req->value.length <= 0)
 		return;
 
-	if (WSPCONFIG(ws)->debug < DEBUG_SENSITIVE &&
-		((req->header.length == 6 && strncasecmp((char*)req->header.data, "Cookie", 6) == 0) ||
-		(req->header.length == 20 && strncasecmp((char*)req->header.data, "X-DTLS-Master-Secret", 20) == 0)))
+	if (WSPCONFIG(ws)->debug < DEBUG_SENSITIVE && header_is_sensitive(&req->header))
 		oclog(ws, LOG_HTTP_DEBUG, "HTTP processing: %.*s: (censored)", (int)req->header.length,
 		      req->header.data);
 	else
@@ -331,9 +348,19 @@ void header_value_check(struct worker_st *ws, struct http_req_st *req)
 		      "Device-type: '%s'", value);
 		break;
 	case HEADER_PLATFORM:
+		if (value_length + 1 > sizeof(req->devplatform)) {
+			req->devplatform[0] = 0;
+			goto cleanup;
+		}
+		memcpy(req->devplatform, value, value_length);
+		req->devplatform[value_length] = 0;
+
 		if (strncasecmp(value, "apple-ios", 9) == 0 ||
 		    strncasecmp(value, "android", 7) == 0) {
 
+			if (strncasecmp(value, "apple-ios", 9) == 0)
+				req->is_ios = 1;
+
 			oclog(ws, LOG_DEBUG,
 			      "Platform: '%s' (mobile)", value);
 			req->is_mobile = 1;
@@ -371,37 +398,32 @@ void header_value_check(struct worker_st *ws, struct http_req_st *req)
 
 		if (strncasecmp(req->user_agent, "Open AnyConnect VPN Agent v", 27) == 0) {
 			unsigned version = atoi(&req->user_agent[27]);
-			if (version <= 3)
+			if (version <= 3) {
+				oclog(ws, LOG_DEBUG, "Detected OpenConnect v3 or older");
 				req->user_agent_type = AGENT_OPENCONNECT_V3;
-			else
+			} else {
+				oclog(ws, LOG_DEBUG, "Detected OpenConnect v4 or newer");
 				req->user_agent_type = AGENT_OPENCONNECT;
+			}
+		} else if (strncasecmp(req->user_agent, "Cisco AnyConnect VPN Agent for Apple", 36) == 0) {
+			oclog(ws, LOG_DEBUG, "Detected Cisco AnyConnect on iOS");
+			req->user_agent_type = AGENT_ANYCONNECT;
+			req->is_ios = 1;
 		} else if (strncasecmp(req->user_agent, "OpenConnect VPN Agent", 21) == 0) {
+			oclog(ws, LOG_DEBUG, "Detected OpenConnect v4 or newer");
 			req->user_agent_type = AGENT_OPENCONNECT;
 		} else if (strncasecmp(req->user_agent, "Cisco AnyConnect", 16) == 0) {
+			oclog(ws, LOG_DEBUG, "Detected Cisco AnyConnect");
 			req->user_agent_type = AGENT_ANYCONNECT;
 		} else if (strncasecmp(req->user_agent, "AnyConnect", 10) == 0) {
+			oclog(ws, LOG_DEBUG, "Detected Cisco AnyConnect");
 			req->user_agent_type = AGENT_ANYCONNECT;
-		}
-
-		switch (req->user_agent_type) {
-			case AGENT_OPENCONNECT_V3:
-				oclog(ws, LOG_DEBUG, "Detected OpenConnect v3 or older");
-				break;
-			case AGENT_OPENCONNECT:
-				oclog(ws, LOG_DEBUG, "Detected OpenConnect v4 or newer");
-				break;
-			case AGENT_ANYCONNECT:
-				oclog(ws, LOG_DEBUG, "Detected Cisco AnyConnect");
-				break;
-			default:
-				oclog(ws, LOG_DEBUG, "Unknown client");
+		} else {
+			oclog(ws, LOG_DEBUG, "Unknown client (%s)", req->user_agent);
 		}
 		break;
 
 	case HEADER_DTLS_CIPHERSUITE:
-		if (req->use_psk || !WSCONFIG(ws)->dtls_legacy)
-			break;
-
 		str = (char *)value;
 
 		p = strstr(str, DTLS_PROTO_INDICATOR);
@@ -415,6 +437,9 @@ void header_value_check(struct worker_st *ws, struct http_req_st *req)
 			}
 		}
 
+		if (req->use_psk || !WSCONFIG(ws)->dtls_legacy)
+			break;
+
 		if (req->selected_ciphersuite) /* if set via HEADER_DTLS12_CIPHERSUITE */
 			break;
 
