@@ -144,7 +144,7 @@ int post_kkdcp_handler(worker_st *ws, unsigned http_ver)
 		return -1;
 	}
 
-	ws_add_score_to_ip(ws, WSCONFIG(ws)->ban_points_kkdcp, 0);
+	ws_add_score_to_ip(ws, WSCONFIG(ws)->ban_points_kkdcp, 0, 0);
 	oclog(ws, LOG_HTTP_DEBUG, "HTTP processing kkdcp framed request: %u bytes", (unsigned)req->body_length);
 
 	length = BUF_SIZE;
@@ -291,9 +291,9 @@ int post_kkdcp_handler(worker_st *ws, unsigned http_ver)
 	ret = 0;
 	goto cleanup;
  fail:
-	cstp_printf(ws,
-		   "HTTP/1.%u 502 %s\r\n\r\n",
-		   http_ver, reason);
+	(void)cstp_printf(ws,
+			 "HTTP/1.%u 502 %s\r\n\r\n",
+			 http_ver, reason);
 	ret = -1;
 
  cleanup:
