@@ -23,7 +23,6 @@
 OCCTL="${OCCTL:-../src/occtl/occtl}"
 SERV="${SERV:-../src/ocserv}"
 srcdir=${srcdir:-.}
-PORT=4574
 PIDFILE=ocserv-pid.$$.tmp
 CLIPID=oc-pid.$$.tmp
 PATH=${PATH}:/usr/sbin
@@ -32,6 +31,8 @@ OUTFILE=traffic.$$.tmp
 
 . `dirname $0`/common.sh
 
+eval "${GETPORT}"
+
 if test -z "${IP}";then
 	echo "no IP tool is present"
 	exit 77
@@ -68,10 +69,14 @@ USERNAME=test
 
 . `dirname $0`/ns.sh
 
+if test -z "${TEST_CONFIG}";then
+	TEST_CONFIG=test-ciphers.config
+fi
+
 # Run servers
-update_config test-ciphers.config
+update_config ${TEST_CONFIG}
 if test "$VERBOSE" = 1;then
-DEBUG="-d 3"
+DEBUG="-d 4"
 fi
 
 ${CMDNS2} ${SERV} -p ${PIDFILE} -f -c ${CONFIG} ${DEBUG} & PID=$!
