@@ -1,4 +1,4 @@
-#!/bin/sh
+#!/bin/bash
 #
 # Copyright (C) 2011-2013 Free Software Foundation, Inc.
 # Copyright 2013 Nikos Mavrogiannopoulos
@@ -23,18 +23,28 @@
 
 builddir=${builddir:-.}
 
-OPENCONNECT=${OPENCONNECT:-/usr/sbin/openconnect}
+OPENCONNECT=${OPENCONNECT:-$(which openconnect)}
 
-if ! test -x ${OPENCONNECT};then
+if test -z "${OPENCONNECT}" || ! test -x ${OPENCONNECT};then
 	echo "You need openconnect to run this test"
-	exit 77
+	exit 1
+fi
+
+if test "${DISABLE_ASAN_BROKEN_TESTS}" = 1;then
+	echo "Disabling worker isolation to enable asan"
+	ISOLATE_WORKERS=false
 fi
 
 if test -z "$NO_NEED_ROOT";then
 	if test "$(id -u)" != "0";then
+		echo "You need to run this script as root"
 		exit 77
 	fi
 else
+	if test "${DISABLE_ASAN_BROKEN_TESTS}" = 1;then
+		echo "Skipping test requiring ldpreload"
+		exit 77
+	fi
 	SOCKDIR="${srcdir}/tmp/sockwrap.$$.tmp"
 	mkdir -p $SOCKDIR
 	export SOCKET_WRAPPER_DIR=$SOCKDIR
@@ -71,7 +81,9 @@ update_config() {
 	       -e 's|@VPNNET6@|'${VPNNET6}'|g' "$file.$$.tmp" \
 	       -e 's|@ROUTE1@|'${ROUTE1}'|g' "$file.$$.tmp" \
 	       -e 's|@ROUTE2@|'${ROUTE2}'|g' "$file.$$.tmp" \
-	       -e 's|@OCCTL_SOCKET@|'${OCCTL_SOCKET}'|g' "$file.$$.tmp"
+	       -e 's|@MATCH_CIPHERS@|'${MATCH_CIPHERS}'|g' "$file.$$.tmp" \
+	       -e 's|@OCCTL_SOCKET@|'${OCCTL_SOCKET}'|g' "$file.$$.tmp" \
+	       -e 's|@LISTEN_NS@|'${LISTEN_NS}'|g' "$file.$$.tmp"
 	CONFIG="$file.$$.tmp"
 }
 
@@ -137,9 +149,9 @@ launch_pam_server() {
 		SR="libsocket_wrapper.so:"
 	fi
 	if test -n "${VERBOSE}" && test "${VERBOSE}" -ge 1;then
-		LD_PRELOAD=libnss_wrapper.so:${SR}libpam_wrapper.so:libuid_wrapper.so PAM_WRAPPER=1 UID_WRAPPER=1 UID_WRAPPER_ROOT=1 $SERV $* &
+		LD_PRELOAD=libnss_wrapper.so:${SR}libpam_wrapper.so:libuid_wrapper.so PAM_WRAPPER=1 UID_WRAPPER=1 UID_WRAPPER_ROOT=1 $PRELOAD_CMD $SERV $* &
 	else
-		LD_PRELOAD=libnss_wrapper.so:${SR}libpam_wrapper.so:libuid_wrapper.so PAM_WRAPPER=1 UID_WRAPPER=1 UID_WRAPPER_ROOT=1 $SERV $* >/dev/null 2>&1 &
+		LD_PRELOAD=libnss_wrapper.so:${SR}libpam_wrapper.so:libuid_wrapper.so PAM_WRAPPER=1 UID_WRAPPER=1 UID_WRAPPER_ROOT=1 $PRELOAD_CMD $SERV $* >/dev/null 2>&1 &
 	fi
 	LOCALPID="$!";
 	unset NSS_WRAPPER_PASSWD
@@ -203,4 +215,48 @@ cleanup() {
 	return $ret
 }
 
+# Check for a utility to list ports.  Both ss and netstat will list
+# ports for normal users, and have similar semantics, so put the
+# command in the caller's PFCMD, or exit, indicating an unsupported
+# test.  Prefer ss from iproute2 over the older netstat.
+have_port_finder() {
+	for file in $(which ss 2> /dev/null) /*bin/ss /usr/*bin/ss /usr/local/*bin/ss;do
+		if test -x "$file";then
+			PFCMD="$file";return 0
+		fi
+	done
+
+	if test -z "$PFCMD";then
+	for file in $(which netstat 2> /dev/null) /bin/netstat /usr/bin/netstat /usr/local/bin/netstat;do
+		if test -x "$file";then
+			PFCMD="$file";return 0
+		fi
+	done
+	fi
+
+	if test -z "$PFCMD";then
+		echo "neither ss nor netstat found"
+		exit 1
+	fi
+}
+
+check_if_port_in_use() {
+	local PORT="$1"
+	local PFCMD; have_port_finder
+	$PFCMD -an|grep "[\:\.]$PORT" >/dev/null 2>&1
+}
+
+# Find a port number not currently in use.
+GETPORT='
+    rc=0
+    unset myrandom
+    while test $rc = 0; do
+        if test -n "$RANDOM"; then myrandom=$(($RANDOM + $RANDOM)); fi
+        if test -z "$myrandom"; then myrandom=$(date +%N | sed s/^0*//); fi
+        if test -z "$myrandom"; then myrandom=0; fi
+        PORT="$(((($$<<15)|$myrandom) % 63001 + 2000))"
+        check_if_port_in_use $PORT;rc=$?
+    done
+'
+
 trap "fail \"Failed to launch the server, aborting test... \"" 10 
